/*
**
**      File : nrd.c 
**      (C)1996-97  Mike Flouris.
**
**      *** Original Code ***
**
**      Description : Code implementing a prototype
**                    device driver for the nrd (Network RAM-Disk)
**                    with a user thread.
**        ( Use tabstop = 4 to view this code right. )
*/

#include "nrd.h"

/*
** ################################
**   The Kernel Include Files ...
** ################################ */
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/blkdev.h>
#include <linux/minix_fs.h>
#include <linux/ext2_fs.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/wait.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/malloc.h>
#include <linux/ioctl.h>
#include <linux/fd.h>
#include <linux/blk.h>
#include <linux/module.h>

#include <linux/hdreg.h>
#include <linux/genhd.h>

#include <linux/net.h>
#include <linux/netdevice.h>
#include <linux/stat.h>
#include <linux/if_arp.h>
#include <linux/in.h>
#include <linux/if.h>
#include <linux/inet.h>
#include <linux/route.h>
#include <net/sock.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>


/*
** #######################################
**   External Functions & Variables ...
** ####################################### */
extern int errno; /* For debugging ... */

/*
** #########################
**     Special Defines
** ######################### */

#ifdef  NRDDEBUG
#define dprintk(args...)    printk(## args)
#else
#define dprintk(args...)
#endif

#define bcopy(s, d, n)  memcpy((d), (s), (n))
    /* sleep for sec milliseconds ... */
#define _S(nr) (1<<((nr)-1))


/*
 *  This struct defines the HD's and their types.
 */
static struct hd_driveid *nrd_ident_info[ 1 ];

/*
** #########################
**      Global Variables
** ######################### */

/* How many times is it open ? */
static int open_nrd = 0;

/* Is nrd initialized ? */
static int user_thread_on = NO;

/* Stuff for communication with user-level thread */
static struct Nrd_User_Info nrd_user_info;
static struct wait_queue *user_sleep;
static struct wait_queue *driver_sleep;
static int user_sleeping = 0, driver_sleeping = 0;

/* Prefferred Block Sizes (for each partition)
   (The block size of the system is 512 bytes)
   This is how many bytes each request will send
*/
static int nrd_blocksizes[ 1 ] = { NRD_BLOCK_SIZE };

/* NRD size in blocks ... */
static int nrd_size_blocks = NRD_HOST_BLOCKS;

/*
***********************************************************
*****                 nrd_request()                   *****
***********************************************************
*/

void nrd_request( void )
{


repeat:
            /* Initialize the request (Kernel macro) ... */
    INIT_REQUEST;

    /*
    ** Checking for incorrect block numbers ...
    *  end_request(0) creates an I/O error ...
    */
    if ( CURRENT->sector < 0 ) {
        dprintk("NRD : Invalid Block Number ...\n");
        CURRENT->errors++;
        end_request(0);
        goto repeat;
    }

    /* See if they called for fun ... */
    if ( CURRENT->nr_sectors == 0 ) {
        dprintk("NRD : Zero Blocks to transfer\n");
        end_request(1);
        goto repeat;
    }

                /*  Okay, is it a Read Request ? */
    if (CURRENT->cmd == READ) {
        
        dprintk("NRD DRIVER : READ Request\n");

        /* Prepare the request Header ... */
                                                /* Read request ... */
        nrd_user_info.req.type = NRD_REQ_BLOCK_READ;
                                                /* How many blocks ??? */
        nrd_user_info.req.blocknum = CURRENT->nr_sectors;
                                                /* Starting Block ??? */
        nrd_user_info.req.stblk = CURRENT->sector;
        
        /* Just for me ... */
        if (CURRENT->current_nr_sectors != CURRENT->nr_sectors)
                dprintk("CURRENT->current_nr_sectors != CURRENT->nr_sectors\n");

        /* Do the really @!#^&*@ hard stuff ... */
        if ( user_sleeping == YES )
            wake_up( &user_sleep );
        driver_sleeping = YES;
        sleep_on( &driver_sleep );
        driver_sleeping = NO;

        /* Maybe needed ? */
        CURRENT->sector++;
        --CURRENT->nr_sectors;
        --CURRENT->current_nr_sectors;

        CURRENT->errors = 0;
        end_request(1);
    }       /* End Read request */


            /* Start Write Request */           
    else if (CURRENT->cmd == WRITE) {

        dprintk("NRD DRIVER : WRITE Request\n");

                                            /* Write request ... */
        nrd_user_info.req.type = NRD_REQ_BLOCK_WRITE;
                                            /* How many blocks ??? */
        nrd_user_info.req.blocknum = CURRENT->nr_sectors;
                                            /* Starting Block ??? */
        nrd_user_info.req.stblk = CURRENT->sector;

            /* Just for me ... */
        if (CURRENT->current_nr_sectors != CURRENT->nr_sectors)
                dprintk("CURRENT->current_nr_sectors != CURRENT->nr_sectors\n");

            /* Do the really hard stuff ... */
        if ( user_sleeping == YES )
            wake_up( &user_sleep );
        driver_sleeping = YES;
        sleep_on( &driver_sleep );
        driver_sleeping = NO;

            /* Maybe needed ? */
        CURRENT->sector++;
        --CURRENT->nr_sectors;
        --CURRENT->current_nr_sectors;

        CURRENT->errors = 0;
        end_request(1);

    } /* Write request */

    else {
        dprintk("NRD : Unknown request ...\n");
        CURRENT->errors++;
        end_request(0);
    }

    dprintk("NRD DRIVER : Finished Request. Returning ...\n");
    goto repeat;

}  /*  End of nrd_request ... */



/*
***********************************************************
*****                  nrd_release()                  *****
***********************************************************
*/


static void nrd_release( struct inode * inode, struct file * file )
{

        /*
        ** Mark the device as closed ...
        */
        if (! (open_nrd < 0) ) 
                open_nrd  = 0;
        else    {
                dprintk("Device already closed ...\n");
                return;
        }
            /* We sync the inode, so that we can forget about it ... */

        return;
}


/*
***********************************************************
*****                   nrd_ioctl()                   *****
***********************************************************
*/

/*
 * IMPORTANT NOTE : Be VERY CAUTIOUS with ioctl commands. The machine
 *          will hang without any reason (with an kernel memory read
 *          access error), without printing any debugging output like
 *          printks. When something like this happens, it's
 *          surely an ioctl error ...
*/


int nrd_ioctl(struct inode * inode, struct file * file,
        unsigned int cmd, unsigned long arg)
{

    int err = 0;
    struct hd_geometry *loc = (struct hd_geometry *) arg;

    dprintk("NRD : nrd_ioctl called\n");

    if ((!inode) || !(inode->i_rdev))
        return -EINVAL;

        switch (cmd) {
                case BLKGETSIZE:   /* Return device size */
                    if (!arg)  return -EINVAL;
                    err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
                    if (err)
                        return err;
                    put_user( (long) nrd_size_blocks , (long *) arg);
                    return 0;
                case BLKFLSBUF:
                    if(!suser())  return -EACCES;
                        fsync_dev(inode->i_rdev);
                        invalidate_buffers(inode->i_rdev);
                    return 0;
                case HDIO_GETGEO:
                    if (!loc)  return -EINVAL;
                    err = verify_area(VERIFY_WRITE, loc, sizeof(*loc));
                    if (err)
                        return err;
                    put_user( 16, (char *) &loc->heads);
                    put_user( (int)(nrd_size_blocks/(16 * 64)), (char *) &loc->sectors);
                    put_user( 64, (short *) &loc->cylinders);
                    put_user( 1, (long *) &loc->start);
                    return 0;
                case BLKRAGET:
                    if (!arg)  return -EINVAL;
                    err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
                    if (err)
                        return err;
                    put_user( 0,(long *) arg);
                    return 0;
                case BLKRASET:
                    if(!suser())  return -EACCES;
                    if(arg > 0xff) return -EINVAL;
                    return 0;
                case BLKRRPART: /* Re-read partition tables */
                    sync_dev( MKDEV(MAJOR_NR, 0) );
                    invalidate_inodes( MKDEV(MAJOR_NR, 0) );
                    invalidate_buffers( MKDEV(MAJOR_NR, 0) );
                    return 0;
                case HDIO_GET_IDENTITY:
                    if (!arg)  return -EINVAL;
                    if (MINOR(inode->i_rdev) != 0 ) return -EINVAL;
                    if (nrd_ident_info[ 0 ] == NULL)  return -ENOMSG;
                    err = verify_area(VERIFY_WRITE, (char *) arg, sizeof(struct hd_driveid));
                    if (err)
                        return err;
                    memcpy_tofs((char *)arg, (char *) nrd_ident_info[0], sizeof(struct hd_driveid));
                    return 0;
                case NRD_IOCTL_USER:
                    if ( driver_sleeping == YES )
                        wake_up( &driver_sleep );
                    user_sleeping = YES;
                    sleep_on( &user_sleep );
                    user_sleeping = NO;
                    if ( nrd_user_info.req.type == NRD_REQ_BLOCK_WRITE) {
                        memcpy_tofs((char *)arg, (char *)&(nrd_user_info.req),
                                    sizeof(IO_Request) );
                        memcpy_tofs((char *)(arg + sizeof(IO_Request)), CURRENT->buffer,
                                    (CURRENT->nr_sectors * SYS_BLOCK_SIZE) );
                    } else      /* READ ... */
                        memcpy_tofs((char *)arg, (char *)&(nrd_user_info.req), sizeof(IO_Request) );
                    return 0;
                case NRD_IOCTL_USER_READ:
                    if (user_thread_on == NO) {
                        if(!suser())  return -EACCES;
                        memcpy_fromfs( (char *)&nrd_size_blocks, (char *)arg, sizeof(int) );
                        nrd_size_blocks = nrd_size_blocks * NRD_HOST_BLOCKS;
                        printk("NRD BLOCKS : %d. \n", nrd_size_blocks);
                        if ( driver_sleeping == YES )
                            wake_up( &driver_sleep );
                        user_thread_on = YES;
                        return 0;
                    }
                    memcpy_fromfs( CURRENT->buffer, (char *) arg,
                                  (CURRENT->nr_sectors * SYS_BLOCK_SIZE) );
                    wake_up( &driver_sleep );
                    return 0;
                case NRD_IOCTL_INIT:
                    if(!suser())  return -EACCES;
                    user_thread_on = NO;
                    return 0;
                default:
                    printk("UNKNOWN IOCTL COMMAND (cmd=%d)\n", cmd);
                    return -EINVAL;
        }
}



/*
***********************************************************
*****                   nrd_open()                    *****
***********************************************************
*/


static int nrd_open( struct inode * inode, struct file * filp )
{


        if ( user_thread_on == NO ) {
            /* This is a very important assignment. It provides the interface
               that will handle the actual reading/writing of blocks ... */
            blk_dev[MAJOR_NR].request_fn = & (nrd_request);

            /* Setting the block size (NRD_BLOCK_SIZE = default) */
            blksize_size[MAJOR_NR] = nrd_blocksizes;

            /* Print this for the user to know that everything
               is okay ... */
            printk("Network Ramdisk driver initialized.\n");

        }   /* if ( user_thread_on == NO )  */

        /*
        ** Open the device ...
        */
        if ( open_nrd >= 0 ) 
                open_nrd++ ;
        else {
                dprintk("nrd_open: open_nrd < 0 ...\n");
            return (ENXIO);
        }

        return 0;
}


/*
***********************************************************
*****               struct  nrd_fops                  *****
***********************************************************
*/

static  struct  file_operations nrd_fops =
{   NULL,       /* int (*lseek)    */
    block_read, /* int (*read)     */
    block_write,/* int (*write)    */
    NULL,       /* int (*readdir)  */
    NULL,       /* int (*select)   */
    nrd_ioctl,  /* int (*ioctl)    */
    NULL,       /* int (*mmap)     */
    nrd_open,   /* int (*open)     */
    nrd_release,/* void (*release) */
    block_fsync,    /* int (*fsync)    */
    NULL,       /* int (*fasync)   */
    NULL,       /* int (*check_media_change) */
    NULL };     /* int (*revalidate) */



/*
***********************************************************
*****                   nrd_init()                    *****
***********************************************************
*/
int 
nrd_init()
{
    
    
    if (register_blkdev( MAJOR_NR, DEVICE_NAME, &nrd_fops) != 0) {
        printk("nrd: unable to get major number %d for Network Ramdisk\n",MAJOR_NR);
        return -EIO;
    }

        /***********************************
        ** Initialize the Network RamDisk
        ************************************/

    /* Set the hd_driveid struct (HD Identity) */
    nrd_ident_info[ 0 ] = (struct hd_driveid *) kmalloc(512,GFP_KERNEL);
    nrd_ident_info[ 0 ]->cyls = 64;
    nrd_ident_info[ 0 ]->heads = 16;
    nrd_ident_info[ 0 ]->sectors = (int) (nrd_size_blocks/(16 * 64));
    nrd_ident_info[ 0 ]->cur_cyls = 64;
    nrd_ident_info[ 0 ]->cur_heads = 16;
    nrd_ident_info[ 0 ]->cur_sectors = (int) (nrd_size_blocks/(16 * 64));
    nrd_ident_info[ 0 ]->buf_size = 512;
    memcpy( nrd_ident_info[ 0 ]->model, "Network RamDisk v2.0", 21);
    nrd_ident_info[ 0 ]->max_multsect = 0;
    nrd_ident_info[ 0 ]->dword_io = 0;
    nrd_ident_info[ 0 ]->capability = 2;
    nrd_ident_info[ 0 ]->multsect = 0;
    nrd_ident_info[ 0 ]->multsect_valid = 0;

    return ( 0 );

}
/* Exiting nrd_init ... */


/*
   These routines are for the module (loadable) version
           (if there can be one ... )
*/
#ifdef MODULE


/*
***********************************************************
*****                init_module()                    *****
***********************************************************
*/

int init_module(void)
{

     int error = nrd_init();
     if (!error)
            printk("NETWORK RAMDISK: Loaded as module.\n");
     return error;

}


/*
***********************************************************
*****                cleanup_module()                 *****
***********************************************************
*/

void cleanup_module(void)
{

        unregister_blkdev( MAJOR_NR, DEVICE_NAME );
        blk_dev[MAJOR_NR].request_fn = 0;

}
#endif /* MODULE */


/* 
***********************************************************
        This is the End , my only friend ...
***********************************************************
*/

