
/*
 * *************************************************************
                     *** Original Code ***

      File        :   nrd_daemon.c
      Author      :   Mike Flouris (C)1997-1998

      Description:
        User level process handling NRD client data requests.
        It sends data and accepts it back over the socket,
        and then sends it back into the kernel device driver.

 * *************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h> 
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "nrd.h"

    /* Leave these untouched. Used only for debugging ... */
#undef  USER_LEVEL_TESTING
#define NET_TESTING

#define NRD_SOCKBUFSIZE  256*1024   /* TCP buffer size ... */

#ifdef  NRDDEBUG
#define dprintf(args...)    printf(## args)
#else
#define dprintf(args...)
#endif

/*
** #########################
**   Function declarations 
** ######################### */
int  send_tcp();
int  recv_tcp();
int  nrd_cleaner ();
void restore_nrd_server();
void * nrd_malloc ();

/*
** #########################
**         Typedefs 
** ######################### */

typedef struct Nrd_Host_Entry {
    char *hostname;         /* Name of the host */
    int  portnum;           /* Port of the server */

    unsigned  num_blocks;        /* Number of blocks it stores */
    unsigned  num_bytes;         /* Number of bytes it stores */

    int  sockfd;            /* Socket file descriptor */

    int  alive;
} Nrd_Host_Entry;

typedef struct Nrd_Server_Info {
    unsigned  server_no;
    Nrd_Host_Entry server_data;
} Nrd_Server_Info;

void    handle_request();
void    nrd_client_sighandler();
int     host_config( void );

unsigned bytes_transferred = 0;
unsigned nrd_size_bytes = 0;
unsigned nrd_size_blocks = 0;

/* Host Table ... */
Nrd_Host_Entry *nrd_servers;

#if defined(NRD_UNRELIABLE_CACHE) || defined(NRD_RELIABLE_PARITY)
    /* This is for the Block Map ... */
typedef struct Nrd_Block_Map_Entry {
    int  server;
    int  block_no;
} Nrd_Block_Map_Entry;
#endif

#ifdef NRD_UNRELIABLE_CACHE
     /* How many blocks in the cache ??? */
unsigned  nrd_blocks_in_cache = 0;

    /* Server Caches ... */
Nrd_Cache *nrd_cache = NULL;

    /* The Block Map ... */
Nrd_Block_Map_Entry *nrd_block_map = NULL;

    /* The server bitmaps (used to manage free
     * space on the servers ... */
char **nrd_server_map = NULL;

#elif defined(NRD_RELIABLE_PARITY)
     /* How many blocks in the cache ??? */
unsigned  nrd_blocks_in_cache[ CACHE_SETS ];

    /* Server Cache Sets (2D cache array) ... */
Nrd_Reliable_Cache **nrd_cache = NULL;

    /* The Block Map ... */
Nrd_Block_Map_Entry *nrd_block_map = NULL;

    /* The server bitmaps (used to manage free
     * space on the servers ... ) */
char **nrd_server_map = NULL;

    /* Variable to count how many times I've sent
     * the write cache over ...
     * Used for the timestamps ... */
unsigned nrd_cache_timestamp = 0;

    /* Timeout in case we are sending the victim set all
     * the time ... */
unsigned nrd_cache_timeout [ CACHE_SETS ];
#endif

    /* Number of NRD servers ... */
int num_nrd_servers = 0;

    /* Number of crashed NRD servers ... */
int nrd_down_server = -1;




    /*
     ***************************************************
     *   Function: main()
     *   Arguments: argc, argv
     *   Description: The main program.
     ***************************************************
     */

int
main(int argc, char **argv) {

    struct  sockaddr_in remoteaddr;
    struct  hostent     *remotehost;
    char    *progname;
    int     i = 0, j = 1, hs = 0;
#if defined(NRD_RELIABLE_PARITY) || defined(NRD_UNRELIABLE_CACHE)
    int     k = 0;
#endif
#ifdef NRD_RELIABLE_PARITY
    int     l = 0, p = 0, s = 0, srv = -1, blk = -1;
#endif
    Nrd_User_Info   nui;
    IO_Request  req;
#ifndef USER_LEVEL_TESTING
    int     nrd_fd;
#endif




        /* Are there any command-line options ? */
    progname = (char *) nrd_malloc( 128 );
    strncpy( progname, argv[0], 128 );
    if ((argc == 1) || (argc >= 64)) {
        fprintf(stderr," Usage : %s [hosthames]\n", progname );
        fflush(stderr);
        exit(2);
    }
    
    num_nrd_servers = argc-1;
    nrd_servers = (Nrd_Host_Entry *) nrd_malloc(num_nrd_servers *
                                                sizeof(Nrd_Host_Entry));
#if defined(NRD_UNRELIABLE_CACHE) || defined(NRD_RELIABLE_PARITY)
    nrd_server_map = (char **) nrd_malloc( num_nrd_servers * sizeof(char *) );
    for (i = 0; i < num_nrd_servers; i++)
        nrd_server_map[i] = (char *) nrd_malloc( NRD_HOST_BLOCKS );
#endif
#ifdef NRD_UNRELIABLE_CACHE
    nrd_cache = (Nrd_Cache *) nrd_malloc( num_nrd_servers * sizeof( Nrd_Cache ) );
    nrd_size_bytes  = num_nrd_servers * NRD_HOST_BYTES;
    nrd_size_blocks = num_nrd_servers * NRD_HOST_BLOCKS;
    nrd_block_map = (Nrd_Block_Map_Entry *) nrd_malloc( nrd_size_blocks *
                                                    sizeof(Nrd_Block_Map_Entry) );
#elif defined(NRD_RELIABLE_PARITY)
    nrd_cache = (Nrd_Reliable_Cache **) nrd_malloc( CACHE_SETS * sizeof(Nrd_Reliable_Cache *) );
    for (k = 0; k < CACHE_SETS; k++)
        nrd_cache[k] = (Nrd_Reliable_Cache *) nrd_malloc( num_nrd_servers *
                                                            sizeof(Nrd_Reliable_Cache) );
    nrd_size_bytes  = (num_nrd_servers-1) * NRD_HOST_BYTES;
    nrd_size_blocks = (num_nrd_servers-1) * NRD_HOST_BLOCKS;
    nrd_block_map = (Nrd_Block_Map_Entry *) nrd_malloc( nrd_size_blocks *
                                                    sizeof(Nrd_Block_Map_Entry) );
#endif
    for (i = 0; i < num_nrd_servers; i++) {
        char    buffer[256];

        if ( argv[i+1] != NULL )
            strncpy( (char *) &buffer, argv[i+1], 256);
        if ( strlen((char *) &buffer) > 0 ) {
            nrd_servers[i].hostname = (char *) nrd_malloc( strlen((char *) &buffer) );
            bzero( nrd_servers[i].hostname, strlen( (char *) &buffer) );
            strcpy( nrd_servers[i].hostname, (char *) &buffer );
        } else {
            fprintf( stderr, "Invalid hostname.\n");
            exit(1);
        }
        nrd_servers[ i ].portnum = NRD_SERVER_PORT;
        nrd_servers[ i ].num_blocks = NRD_HOST_BLOCKS;
        nrd_servers[ i ].num_bytes  = NRD_HOST_BYTES;
        printf("Server [%d] on %s:%d (%d blocks) OK.\n", i, nrd_servers[i].hostname,
            nrd_servers[i].portnum, nrd_servers[ i ].num_blocks);
    }

#ifdef NRD_RELIABLE_PARITY
    if ( num_nrd_servers < 3 ) {
        fprintf( stderr,"Parity Reliable NRD requires at least 3 NRD Servers.\n");
        exit(1);
    }
    /* Mark the Cache empty ... */
    for (i = 0; i < num_nrd_servers; i++) {
        for (s = 0; s < CACHE_SETS; s++) {
            for (j = 0; j < CACHE_SIZE_BLOCKS; j++) {
                nrd_cache[s][i].p_block_index[j] = -1;
                nrd_cache[s][i].v_block_index[j] = -1;
            }
            for (j = 0; j < CACHE_SIZE_BYTES; j++)
                nrd_cache[s][i].block_data[j] = 0;
        }
    }
    /* Zero nrd_blocks_in_cache & nrd_cache_timeout */
    for (s = 0; s < CACHE_SETS; s++) {
        nrd_blocks_in_cache[s] = 0;
        nrd_cache_timeout[s] = 0;
    }

#elif defined(NRD_UNRELIABLE_CACHE)
    /* Mark the Cache empty ... */
    for (i = 0; i < num_nrd_servers; i++) {
        for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
            nrd_cache[i].block_index[j] = -1;
        for (j = 0; j < CACHE_SIZE_BYTES; j++)
            nrd_cache[i].block_data[j] = 0;
    }
#endif
#if defined(NRD_UNRELIABLE_CACHE) || defined(NRD_RELIABLE_PARITY)
    /* Mark the Bitmaps empty ... */
    for (i = 0; i < num_nrd_servers; i++)
        for (j = 0; j < NRD_HOST_BLOCKS; j++)
                nrd_server_map[i][j] = NRD_PBLOCK_EMPTY;
    /* Mark the Block Map empty ... */
    for (i = 0; i < nrd_size_blocks; i++) {
        nrd_block_map[ i ].block_no = -1;
            /* -1 = invalid value for server ... */
        nrd_block_map[ i ].server = -1;
    }
    printf("Initialized maps OK. (%d blocks)\n", nrd_size_blocks);
#endif

                        /* Connect to the servers ... */
    for ( hs = 0; hs < num_nrd_servers; hs++) {

        printf("Connecting to %s ...\n", nrd_servers[hs].hostname);
        if ((remotehost = gethostbyname(nrd_servers[hs].hostname)) == NULL) {
            perror("gethostbyname ");
            return(-1);
        }

        bzero((char *) &remoteaddr, sizeof(struct sockaddr_in));
        bcopy(remotehost->h_addr,(char *)&remoteaddr.sin_addr,remotehost->h_length);
        remoteaddr.sin_port     =   htons(NRD_SERVER_PORT);
        remoteaddr.sin_family   =   remotehost->h_addrtype;
    
        /*
        ** Creating the socket ... */
        if((nrd_servers[hs].sockfd = socket(remotehost->h_addrtype, SOCK_STREAM, 0)) < 0) {
            perror("socket_create ");
            exit(1);
        }

        /*
        ** Setting the non-blocking socket options ... */
        if(setsockopt(nrd_servers[hs].sockfd, IPPROTO_TCP, TCP_NODELAY,
                      (char *)&j, sizeof(int)) == -1) {
            perror("Seting TCP_NODELAY ");
            exit(1);
        }

#ifdef NET_TESTING
        /*
         ** Connecting the socket ... */
        if (connect(nrd_servers[hs].sockfd, (struct sockaddr *)&remoteaddr,
                    sizeof( struct sockaddr )) ) {
            perror("socket connect ");
            return (-1);
        } else
#endif
            nrd_servers[hs].alive = 1;

            /* Hangs machine if no space available ... */
#if defined(NET_TESTING) && defined(SET_TCP_BUFFER)
        j = NRD_SOCKBUFSIZE;
        if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_SNDBUF,
                      (char *)&j, sizeof(int)) == -1) {
                perror("Seting SND_BUF failed ");
                exit(1);
            }
        j = NRD_SOCKBUFSIZE;
        if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_RCVBUF,
                      (char *)&j, sizeof(int)) == -1) {
                perror("Seting RCV_BUF failed ");
                exit(1);
         }
#endif


    } /* for ... */

#ifndef USER_LEVEL_TESTING
        /* Open the device driver descriptor ... */
    if ( (nrd_fd = open("/dev/nrd", O_RDWR, S_IRWXU)) == -1) {
        perror("Opening nrd");
        exit(1);
    }

        /* Initialize the device driver ... */
    if (ioctl(nrd_fd, NRD_IOCTL_INIT, NULL ) != 0) {
        perror("Initial ioctl (NRD_IOCTL_INIT)");
        exit(1);
    }

#ifdef NRD_RELIABLE_PARITY
        /* Tell the device driver how many servers are there ... */
    i = num_nrd_servers-1;
    if (ioctl(nrd_fd, NRD_IOCTL_USER_READ, (char *) &i) != 0) {
        perror("Initial ioctl (NRD_IOCTL_USER_READ)");
        exit(1);
    }
#else
        /* Tell the device driver how many servers are there ... */
    if (ioctl(nrd_fd, NRD_IOCTL_USER_READ, (char *) &num_nrd_servers) != 0) {
        perror("Initial ioctl (NRD_IOCTL_USER_READ)");
        exit(1);
    }
#endif

#endif

    /* NRD Signal Handler to perform necessary cleanups */
    signal(SIGINT, nrd_client_sighandler);

        /* Start doing the requests ... */
    while ( 1 ) {

#ifndef USER_LEVEL_TESTING

        /* Get into the device driver & wait for a request ... */
        if (ioctl(nrd_fd, NRD_IOCTL_USER, (char *) &nui) != 0) {
            perror("ioctl (NRD_IOCTL_USER)");
            exit(1);
        }

#endif


#ifdef NRD_UNRELIABLE_NOCACHE
                    /* Which server is it ??? */
        hs = nui.req.stblk / NRD_HOST_BLOCKS;

                    /* Prepare the header ... */
        req.type = htonl( nui.req.type );
        req.blocknum = htonl( nui.req.blocknum );
        req.stblk = htonl( nui.req.stblk % NRD_HOST_BLOCKS );
#endif
                    /* Is it a Read request ??? */
        if ( nui.req.type == NRD_REQ_BLOCK_READ ) {

#ifdef NRD_UNRELIABLE_NOCACHE
                    /* Send the Read Request */
            dprintf("NRD USER : Read Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);
#ifdef NET_TESTING
            send_tcp( nrd_servers[hs].sockfd, &req, (3*sizeof(int)) );

                    /* Read the data */
            recv_tcp( nrd_servers[hs].sockfd, (char *) nui.data,
                      (nui.req.blocknum * SYS_BLOCK_SIZE));
#endif

            /* OKAY, we have a block cache. Hang on ... */
#elif defined( NRD_UNRELIABLE_CACHE )
            int in_cache = 0;
            dprintf("NRD USER : Read Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);

            /* Look if the blocks are in the cache ... */
            if (nrd_blocks_in_cache != 0) {
                for (i = 0; i < num_nrd_servers; i++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        if (nrd_cache[i].block_index[j] == nui.req.stblk) {
                            dprintf("BLOCK FOUND IN CACHE ...\n");

                            /* They are here : Copy them back ... */
                            bcopy(  (char *) &(nrd_cache[i].block_data[j * SYS_BLOCK_SIZE]),
                                    (char *)&(nui.data[0]), (nui.req.blocknum * SYS_BLOCK_SIZE) );
                            in_cache = 1;
                            break;
                        }
            }

            if ( !in_cache ) {
                /* Which server has them ??? */
                hs = nrd_block_map[ nui.req.stblk ].server;

                /* Safety/Sanity checks ... */
                if ( hs < 0) {
                    fprintf( stderr, "Read request for unwritten block !\n");

#ifndef USER_LEVEL_TESTING
                        /* Put the data back into the device driver ... */
                    dprintf("NRD USER : Calling NRD_IOCTL_USER_READ \n");
                    if (ioctl(nrd_fd, NRD_IOCTL_USER_READ, (char *) &(nui.data[0]) ) != 0) {
                        perror("ioctl (NRD_IOCTL_USER_READ)");
                        exit(1);
                    }
#endif
                    continue;
                }
                if ( hs != nrd_block_map[ nui.req.stblk + 1 ].server ) {
                    fprintf( stderr, "Sequential blocks are in different servers !\n");
                    exit(1);
                }
                if ( nrd_block_map[ nui.req.stblk ].block_no + 1 !=
                    nrd_block_map[ nui.req.stblk + 1 ].block_no ) {
                    fprintf( stderr, "Sequential blocks have not sequential numbers !\n");
                    exit(1);
                }

                /* Prepare the header ... */
                req.type = htonl( nui.req.type );
                req.blocknum = htonl( nui.req.blocknum );
                req.stblk = htonl( nrd_block_map[ nui.req.stblk ].block_no );

#ifdef NET_TESTING
                /* Get it from there ... */
                    /* Send the read request header ... */
                send_tcp( nrd_servers[hs].sockfd, &req, sizeof(IO_Request) );

                    /* Read the data */
                recv_tcp( nrd_servers[hs].sockfd, (char *) nui.data,
                        (nui.req.blocknum * SYS_BLOCK_SIZE));
#endif

            } /* if ( !in_cache) ... */

#elif defined( NRD_RELIABLE_PARITY )
            int in_cache = 0;
            dprintf("NRD USER : Read Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);

            /* Look if the blocks are in the cache ... */
            for (s = 0; s < CACHE_SETS; s++)
                if (nrd_blocks_in_cache[s] != 0)
                for (i = 0; i < num_nrd_servers-1; i++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        if (nrd_cache[s][i].v_block_index[j] == nui.req.stblk) {
                            dprintf("BLOCK FOUND IN CACHE ...\n");

                            /* They are here : Copy them back ... */
                            bcopy(  (char *) &(nrd_cache[s][i].block_data[j * SYS_BLOCK_SIZE]),
                                    (char *) &(nui.data[0]), (nui.req.blocknum * SYS_BLOCK_SIZE) );
                            in_cache = 1;
                            goto rep_found_in_cache;
                        }
rep_found_in_cache:

                /* If the blocks are not in the cache ... */
            if ( ! in_cache ) {

                /* Which server has them ??? */
                hs = nrd_block_map[ nui.req.stblk ].server;

                /* Check if the server is alive ... */
                if ((hs != -1) && (nrd_servers[hs].alive == 0)) {
                    char *recov_data;
                    dprintf( stderr, "NRD_READ : Server %d is dead ...\n", hs);
                    dprintf( stderr, "NRD_READ : Must reconstruct the block from parity ...\n" );

                        /* Read all the corresponding blocks from all the
                         * other servers ... */

                    /* Prepare the header ... */
                    req.type = htonl( nui.req.type );
                    req.blocknum = htonl( nui.req.blocknum );
                    req.stblk = htonl( nrd_block_map[ nui.req.stblk ].block_no );

                    bzero( (char *) nui.data, NRD_BLOCK_SIZE );
                    recov_data = (char *) nrd_malloc( NRD_BLOCK_SIZE );
                    for (i = 0; i < num_nrd_servers; i++)
                        if (nrd_servers[i].alive != 0) {

#ifdef NET_TESTING
                            /* Get it from there ... */
                                /* Send the read request header ... */
                            if (send_tcp( nrd_servers[i].sockfd, &req, (3*sizeof(int)) ) == 0) {
                                fprintf( stderr, "More than one servers crashed !\n");
                                fprintf( stderr, "Data is lost. Cannot continue ...\n");
                                exit( 101 );
                            }

                                /* Read the data */
                            if ( recv_tcp( nrd_servers[i].sockfd, (char *) recov_data,
                                                NRD_BLOCK_SIZE) == 0 ) {
                                fprintf( stderr, "More than one servers crashed !\n");
                                fprintf( stderr, "Data is lost. Cannot continue ...\n");
                                exit( 102 );
                            }
#endif
                                /* Calculate the block from the parity ... */
                            for (j = 0; j < NRD_BLOCK_SIZE; j++)
                                nui.data[j] ^= recov_data[j];
                        }

                    free( recov_data );
#ifdef NET_TESTING
                    /* Try to reconnect to the crashed server ... */
                    if ((remotehost = gethostbyname(nrd_servers[hs].hostname)) == NULL) {
                        perror("gethostbyname ");
                        return(-1);
                    }

                    bzero((char *) &remoteaddr, sizeof(struct sockaddr_in));
                    bcopy(remotehost->h_addr,(char *)&remoteaddr.sin_addr,remotehost->h_length);
                    remoteaddr.sin_port     =   htons(NRD_SERVER_PORT);
                    remoteaddr.sin_family   =   remotehost->h_addrtype;
    
                    /*
                    ** Creating the socket ... */
                    if((nrd_servers[hs].sockfd = socket(remotehost->h_addrtype, SOCK_STREAM, 0)) < 0) {
                        perror("socket_create ");
                        exit(1);
                    }

                    /*
                    ** Setting the non-blocking socket options ... */
                    if(setsockopt(nrd_servers[hs].sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&j,
                                  sizeof(int)) == -1) {
                        perror("Seting TCP_NODELAY ");
                        exit(1);
                    }

                    /*
                    ** Connecting the socket ... */
                    if (connect(nrd_servers[hs].sockfd, (struct sockaddr *)&remoteaddr,
                                sizeof( struct sockaddr )) == 0 )
                        restore_nrd_server( hs );
                    else
                        goto put_data_in_kernel;

                        /* Hangs machine if no space available ... */
#ifdef SET_TCP_BUFFER
                    j = NRD_SOCKBUFSIZE;
                    if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_SNDBUF, (char *)&j,
                                  sizeof(int)) == -1) {
                            perror("Seting SND_BUF failed ");
                            exit(1);
                        }
                    j = NRD_SOCKBUFSIZE;
                    if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_RCVBUF, (char *)&j,
                                  sizeof(int)) == -1) {
                            perror("Seting RCV_BUF failed ");
                            exit(1);
                    }
#endif
#endif

                    goto put_data_in_kernel;
                }

                /* Safety/Sanity checks ... */
                if (hs < 0) {
                    dprintf( "Read request for unwritten block !\n");
                    goto put_data_in_kernel;
                }

                if ( hs != nrd_block_map[ nui.req.stblk + 1 ].server ) {
                    fprintf( stderr, "Sequential blocks are in different servers !\n");
                    fprintf( stderr, "Block[%d].server = %d -- Block[%d].server = %d\n",
                                    nui.req.stblk, nrd_block_map[nui.req.stblk].server,
                                    nui.req.stblk+1, nrd_block_map[nui.req.stblk+1].server );
                    exit(1);
                }

                if ( nrd_block_map[ nui.req.stblk ].block_no + 1 !=
                    nrd_block_map[ nui.req.stblk + 1 ].block_no ) {
                    fprintf( stderr, "Sequential blocks have not sequential numbers !\n");
                    fprintf( stderr, "Block[%d].block_no = %d -- Block[%d].block_no = %d\n",
                                    nui.req.stblk, nrd_block_map[nui.req.stblk].block_no,
                                    nui.req.stblk+1, nrd_block_map[nui.req.stblk+1].block_no );
                    exit(1);
                }

                /* Prepare the header ... */
                req.type = htonl( nui.req.type );
                req.blocknum = htonl( nui.req.blocknum );
                req.stblk = htonl( nrd_block_map[ nui.req.stblk ].block_no );

#ifdef NET_TESTING
                /* Get it from there ... */
                    /* Send the read request header ... */
                if (send_tcp( nrd_servers[hs].sockfd, &req, (3*sizeof(int)) ) == 0) {
                    fprintf( stderr, "Server %d not responding ...\n", hs);
                    if ((nrd_down_server != -1) && (nrd_down_server != hs)) {
                        fprintf( stderr, "More than one servers crashed !\n");
                        fprintf( stderr, "Data is lost. Exiting ...\n");
                        exit(121);
                    } else
                        nrd_down_server = hs;
                    nrd_servers[hs].alive = 0;
                    close(nrd_servers[hs].sockfd);
                }

                    /* Read the data */
                if ( recv_tcp( nrd_servers[hs].sockfd, (char *) nui.data,
                              (nui.req.blocknum * SYS_BLOCK_SIZE)) == 0 ) {
                    fprintf( stderr, "Server %d not responding ...\n", hs);
                    if ((nrd_down_server != -1) && (nrd_down_server != hs)) {
                        fprintf( stderr, "More than one servers crashed !\n");
                        fprintf( stderr, "Data is lost. Exiting ...\n");
                        exit(122);
                    } else
                        nrd_down_server = hs;
                    nrd_servers[hs].alive = 0;
                    close(nrd_servers[hs].sockfd);
                }
#endif

                if ((hs != -1) && (nrd_servers[hs].alive == 0)) {
                    char *recov_data;
                    dprintf( "NRD_READ : Server %d is dead ...\n", hs);
                    dprintf( "NRD_READ : Must reconstruct the block from parity ...\n" );

                        /* Read all the corresponding blocks from all the
                         * other servers ... */

                    /* Prepare the header ... */
                    req.type = htonl( nui.req.type );
                    req.blocknum = htonl( nui.req.blocknum );
                    req.stblk = htonl( nrd_block_map[ nui.req.stblk ].block_no );

                    bzero( (char *) nui.data, NRD_BLOCK_SIZE );
                    recov_data = (char *) nrd_malloc( NRD_BLOCK_SIZE );
                    for (i = 0; i < num_nrd_servers; i++)
                        if (nrd_servers[i].alive != 0) {

#ifdef NET_TESTING
                            /* Get it from there ... */
                                /* Send the read request header ... */
                            if (send_tcp( nrd_servers[i].sockfd, &req, (3*sizeof(int)) ) == 0) {
                                fprintf( stderr, "More than one servers crashed !\n");
                                fprintf( stderr, "Data is lost. Cannot continue ...\n");
                                exit( 103 );
                            }

                                /* Read the data */
                            if ( recv_tcp( nrd_servers[i].sockfd, (char *) recov_data,
                                                NRD_BLOCK_SIZE) == 0 ) {
                                fprintf( stderr, "More than one servers crashed !\n");
                                fprintf( stderr, "Data is lost. Cannot continue ...\n");
                                exit( 104 );
                            }
#endif
                                /* Calculate the block from the parity ... */
                            for (j = 0; j < NRD_BLOCK_SIZE; j++)
                                nui.data[j] ^= recov_data[j];
                        }

                    free( recov_data );
#ifdef NET_TESTING
                    /* Try to reconnect to the crashed server ... */
                    if ((remotehost = gethostbyname(nrd_servers[hs].hostname)) == NULL) {
                        perror("gethostbyname ");
                        return(-1);
                    }

                    bzero((char *) &remoteaddr, sizeof(struct sockaddr_in));
                    bcopy(remotehost->h_addr,(char *)&remoteaddr.sin_addr,remotehost->h_length);
                    remoteaddr.sin_port     =   htons(NRD_SERVER_PORT);
                    remoteaddr.sin_family   =   remotehost->h_addrtype;
    
                    /*
                    ** Creating the socket ... */
                    if((nrd_servers[hs].sockfd = socket(remotehost->h_addrtype, SOCK_STREAM, 0)) < 0) {
                        perror("socket_create ");
                        exit(1);
                    }

                    /*
                    ** Setting the non-blocking socket options ... */
                    if(setsockopt(nrd_servers[hs].sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&j,
                                  sizeof(int)) == -1) {
                        perror("Seting TCP_NODELAY ");
                        exit(1);
                    }

                    /*
                    ** Connecting the socket ... */
                    if (connect(nrd_servers[hs].sockfd, (struct sockaddr *)&remoteaddr,
                                sizeof( struct sockaddr )) == 0 )
                        restore_nrd_server( hs );
                    else
                        goto put_data_in_kernel;

                        /* Hangs machine if no space available ... */
#ifdef SET_TCP_BUFFER
                    j = NRD_SOCKBUFSIZE;
                    if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_SNDBUF, (char *)&j,
                                  sizeof(int)) == -1) {
                            perror("Seting SND_BUF failed ");
                            exit(1);
                        }
                    j = NRD_SOCKBUFSIZE;
                    if(setsockopt(nrd_servers[hs].sockfd, SOL_SOCKET, SO_RCVBUF, (char *)&j,
                                  sizeof(int)) == -1) {
                            perror("Seting RCV_BUF failed ");
                            exit(1);
                    }
#endif
#endif

                }
            }   /* if ( ! in_cache ) ... */
#else
            "YOU MUST EDIT nrd.h TO CONFIGURE the NETWORK RAMDISK"
#endif

#ifdef NRD_RELIABLE_PARITY
put_data_in_kernel:
#endif

#ifndef USER_LEVEL_TESTING
                    /* Put the data back into the device driver ... */
            dprintf("NRD USER : Calling NRD_IOCTL_USER_READ \n");
            if (ioctl(nrd_fd, NRD_IOCTL_USER_READ, (char *) &(nui.data[0]) ) != 0) {
                perror("ioctl (NRD_IOCTL_USER_READ)");
                exit(1);
            }
#endif

        } /* End of Read Request ... */

                /* Is it a Write request ??? */
        else if ( nui.req.type == NRD_REQ_BLOCK_WRITE ) {

#ifdef NRD_UNRELIABLE_NOCACHE
                /* Send the Write Request */
            dprintf("NRD USER : Write Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);

#ifdef NET_TESTING
            send_tcp( nrd_servers[hs].sockfd, &req, (3*sizeof(int)) );

                /* Write the data ... */
            send_tcp( nrd_servers[hs].sockfd, (char *) nui.data,
                      (nui.req.blocknum * SYS_BLOCK_SIZE));
#endif

            /* OKAY, we have a block cache. Hang on ... */
#elif defined( NRD_UNRELIABLE_CACHE )
            dprintf("NRD USER : Write Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);

            /* Are the blocks already in the cache ??? */
                /* Search ... */
            for (i = 0; i < num_nrd_servers; i++)
                for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                    if ((nrd_cache[i].block_index[j] == nui.req.stblk) &&
                        (nrd_cache[i].block_index[j+1] == nui.req.stblk+1))
                            goto un_found_in_cache;
un_found_in_cache:

                /* Are they ? */
            if ((i < num_nrd_servers) && (j < CACHE_SIZE_BLOCKS)) {
                for (k = 0; k < nui.req.blocknum; k++) {
                    nrd_cache[i].block_index[j+k] = nui.req.stblk + k;
                    dprintf("Block %d is already in the cache ...\n", nui.req.stblk + k);
                }
                bcopy( (char *)(nui.data), (char *) &(nrd_cache[i].block_data[j * SYS_BLOCK_SIZE]),
                            (nui.req.blocknum * SYS_BLOCK_SIZE) );
                continue;
            }

            /* Is there any space in the cache ??? */
            if ( nrd_blocks_in_cache < ((CACHE_SIZE_BLOCKS * num_nrd_servers) / 2) ) {
                dprintf("********** CACHE NOT FULL ***********\n");

                /* There is some space left. Find the place ... */
            for (i = 0; i < num_nrd_servers; i++)
                for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                    if ((nrd_cache[i].block_index[j] == -1) &&
                        (nrd_cache[i].block_index[j+1] == -1))
                            goto un_find_a_place;
un_find_a_place:


                /* Write the blocks there ... */
                for (k = 0; k < nui.req.blocknum; k++)
                    nrd_cache[i].block_index[j+k] = nui.req.stblk + k;
                bcopy( (char *)(nui.data), (char *) &(nrd_cache[i].block_data[j * SYS_BLOCK_SIZE]),
                            (nui.req.blocknum * SYS_BLOCK_SIZE) );
                nrd_blocks_in_cache += (nui.req.blocknum / 2);
            } /* if (blocks_in_cache ... */

            /* The cache is full, we have to send it over ... */
            if ( nrd_blocks_in_cache >= ((CACHE_SIZE_BLOCKS * num_nrd_servers) / 2) ) {
                dprintf("********** CACHE FULL : Send it over ...\n");

                /* Make the proper block mappings ... */
                for (i = 0; i < num_nrd_servers; i++) {
                    unsigned servblock = 0;

                    for (j = 0; j < CACHE_SIZE_BLOCKS; j+=2) {
                        unsigned server = 0, block_no = 0;

                        /* Find out if the blocks are already written ...*/
                        server = nrd_block_map[ nrd_cache[i].block_index[j] ].server;
                        block_no = nrd_block_map[ nrd_cache[i].block_index[j] ].block_no;
                        if ((server != -1) && (block_no >= 0)) {
                            /* Mark them as empty ... */
                            nrd_server_map[server][block_no] = NRD_PBLOCK_EMPTY;
                            nrd_server_map[server][block_no+1] = NRD_PBLOCK_EMPTY;
                        }

                        /* Find an empty block on the new server ... */
                        for (k = servblock; k < NRD_HOST_BLOCKS; k+=2)
                            if ((nrd_server_map[i][k] == NRD_PBLOCK_EMPTY) &&
                                    (nrd_server_map[i][k+1] == NRD_PBLOCK_EMPTY))
                                break;
                        /* Server is full ... */
                        if ( k >= NRD_HOST_BLOCKS ) {
                            fprintf( stderr, "Server is full ...\n");
                            fprintf( stderr, "This should be checked by the FS !\n");
                            continue;
                        }
                        servblock = k;

                        /* Update the Block Map ... */
                        nrd_block_map[ nrd_cache[i].block_index[j] ].server = i;
                        nrd_block_map[ nrd_cache[i].block_index[j] ].block_no = k;
                        nrd_block_map[ nrd_cache[i].block_index[j+1] ].server = i;
                        nrd_block_map[ nrd_cache[i].block_index[j+1] ].block_no = k+1;

                        /* Mark them as full ... */
                        nrd_server_map[i][k] = NRD_PBLOCK_FULL;
                        nrd_server_map[i][k+1] = NRD_PBLOCK_FULL;

                        /* Map the real indexes on the block_index */
                        nrd_cache[i].block_index[j] = k;
                        nrd_cache[i].block_index[j+1] = k+1;
                }
                }

                /* Prepare the cached write req header ... */
                req.type = htonl( NRD_REQ_CACHE_WRITE );
                req.blocknum = htonl( 0 );
                req.stblk = htonl( 0 );

#ifdef NET_TESTING
                /* Send cache to all the servers ... */
                for (k = 0; k < num_nrd_servers; k++) {

                    /* Send the header ... */
                    dprintf("********** SENDING HEADER WRITE_CACHE ...\n");
                    send_tcp( nrd_servers[k].sockfd, &req, sizeof(IO_Request) );

                    /* Transform the block indexes into network byte order */
                    for (i = 0;i < CACHE_SIZE_BLOCKS;i++)
                        nrd_cache[k].block_index[i] = htonl( nrd_cache[k].block_index[i] );

                    /* Write the data to all the servers ... */
                    dprintf("********** SENDING CACHE DATA ...\n");
                    send_tcp( nrd_servers[k].sockfd, (char *) &(nrd_cache[k]), sizeof(Nrd_Cache) );

                } /* for all the servers ... */
#endif

                /* Empty the Cache now ... */
                dprintf("********** EMPTY THE CACHE ...\n");
                nrd_blocks_in_cache = 0;
                for (i = 0; i < num_nrd_servers; i++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        nrd_cache[i].block_index[j] = -1;

            }

#elif defined( NRD_RELIABLE_PARITY )
            dprintf("NRD USER : Write Request, Block %d, %d blocks\n", nui.req.stblk,
                    nui.req.blocknum);

            /* Are the blocks already in the cache ??? */
                /* Search ... */
            for (s = 0; s < CACHE_SETS; s++)
                for (i = 0; i < num_nrd_servers-1; i++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        if ((nrd_cache[s][i].v_block_index[j] == nui.req.stblk) &&
                            (nrd_cache[s][i].v_block_index[j+1] == nui.req.stblk+1))
                                goto re_found_in_cache;
re_found_in_cache:

                /* Are they ? */
            if ((s < CACHE_SETS) && (i < num_nrd_servers-1) && (j < CACHE_SIZE_BLOCKS)) {
                dprintf("Block %d is already in the cache ...\n", nui.req.stblk + k);
                bcopy( (char *) &(nui.data[0]),
                            (char *) &(nrd_cache[s][i].block_data[j * SYS_BLOCK_SIZE]),
                            (nui.req.blocknum * SYS_BLOCK_SIZE) );
                goto end_write_req;
            }

            /* Is there any cache set full ??? */
            for (s = 0; s < CACHE_SETS; s++)
                if (nrd_blocks_in_cache[s] >= ((CACHE_SIZE_BLOCKS * (num_nrd_servers-1)) / 2) )
                    break;

            if ( s >= CACHE_SETS ) {

                /* ADAPTIVE PLACEMENT METHOD
                 * Try to place the blocks near their sequential
                 * predecessors ... */

                    /* Find out if all sets are empty ... */
                i = 1;
                for (s = 0; s < CACHE_SETS; s++)
                    if (nrd_blocks_in_cache[s] != 0)
                        i = 0;
                if (i == 1) {
                    if (nrd_servers[i].alive != 0) {
                        s = 0; i = 0; j = 0;
                    } else {
                        s = 0; i = 1; j = 0;
                    }
                    goto adaptive_end;
                }

                    /* There is at least one non-empty set. Search for the
                     * predecessors of the blocks ... */
                for (s = 0; s < CACHE_SETS; s++)
                    if (nrd_blocks_in_cache[s] != 0)
                        for (i = 0; i < num_nrd_servers-1; i++)
                            if (nrd_servers[i].alive != 0)
                            for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                                if ((nrd_cache[s][i].v_block_index[j] != -1) &&
                                    (nrd_cache[s][i].v_block_index[j] == nui.req.stblk-1) )
                                    goto pred_findout;
pred_findout:

                    /* The predecessors of the block are in. Find out
                     * where they are ... */
                if ((s < CACHE_SETS) && (i < num_nrd_servers-1) && (j < CACHE_SIZE_BLOCKS)) {
                    int svi = 0, svj = 0;

                    svi = i; svj = j;
                        /* Approximate the next available position
                         * for the block ... */
                    for (j = svj-1; j < CACHE_SIZE_BLOCKS; j++)
                        for (i = 0; i < num_nrd_servers-1; i++)
                            if ((nrd_servers[i].alive != 0) &&
                               (nrd_cache[s][i].v_block_index[j] == -1)) {
                                int flag = 1;
                                for (k = 0;k < nui.req.blocknum;k++)
                                    if ((j+k < CACHE_SIZE_BLOCKS) &&
                                            (nrd_cache[s][i].v_block_index[j+k] != -1))
                                        flag = 0;
                                if (flag)   goto found_cont_free;
                            }
found_cont_free:

                    /* There is no space in this cache set ???
                     * Then proceed like the predecessors of
                     * the block are not in ... */
                    if ((s < CACHE_SETS) && (i < num_nrd_servers-1) &&
                            (j < CACHE_SIZE_BLOCKS))
                        goto adaptive_end;

                }
                    /* The predecessors of the block are not in. */

                    /* Find out if there exists one empty set. */
                    for (s = 0; s < CACHE_SETS; s++)
                        if (nrd_blocks_in_cache[s] == 0)
                            break;

                        /* It exists. Place the blocks in that set ... */
                    if ( s < CACHE_SETS ) {
                        i = 0; j = 0;
                        goto adaptive_end;

                        /* Find a place in the Victim or time-out set ... */
                    } else {
                            
                        /* Find out if any set has timed out ... */
                        for (s = 0; s < CACHE_SETS-1; s++)
                            if (nrd_cache_timeout[s] > CACHE_TIMEOUT)
                                break;

                            /* Yes we have one. Find a place in that set ... */
                        if ( s < CACHE_SETS-1 ) {

                            if (nrd_blocks_in_cache[s] != 0) {
                                for (i = 0; i < num_nrd_servers-1; i++)
                                    if (nrd_servers[i].alive != 0)
                                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                                        if (nrd_cache[s][i].v_block_index[j] == -1) {
                                            int flag = 1;

                                            for (k = 0;k < nui.req.blocknum;k++)
                                                if ((j+k < CACHE_SIZE_BLOCKS) &&
                                                        (nrd_cache[s][i].v_block_index[j+k] != -1))
                                                    flag = 0;
                                                /* Place the blocks in the timed-out set ... */
                                            if (flag) {
                                                goto adaptive_end;
                                            }
                                        }
                            } else {
                                fprintf(stderr, "Timed-out set should not be empty. Exiting ...\n");
                                exit(1);
                            }

                            /* Find a place in the Victim set ... */
                        } else {
                            if (nrd_blocks_in_cache[ CACHE_SETS-1 ] != 0) {
                                for (i = 0; i < num_nrd_servers-1; i++)
                                    if (nrd_servers[i].alive != 0)
                                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                                        if (nrd_cache[ CACHE_SETS-1 ][i].v_block_index[j] == -1) {
                                            int flag = 1;

                                            for (k = 0;k < nui.req.blocknum;k++)
                                                if ((j+k < CACHE_SIZE_BLOCKS) &&
                                                        (nrd_cache[ CACHE_SETS-1 ][i].v_block_index[j+k] != -1))
                                                    flag = 0;
                                                /* Place the blocks in the victim set ... */
                                            if (flag) {
                                                s = CACHE_SETS-1;
                                                goto adaptive_end;
                                            }
                                        }
                            } else {
                                fprintf(stderr, "Victim set should not be empty. Exiting ...\n");
                                exit(1);
                            }

                        } /* else ... */

                    } /* else ... */
adaptive_end:

                /* Write the blocks there ... */
                for (k = 0; k < nui.req.blocknum; k++)
                    nrd_cache[s][i].v_block_index[j+k] = nui.req.stblk + k;
                bcopy( (char *) nui.data, (char *) &(nrd_cache[s][i].block_data[j * SYS_BLOCK_SIZE]),
                            (nui.req.blocknum * SYS_BLOCK_SIZE) );
                nrd_blocks_in_cache[s]++;

            } /* if ( s >= CACHE_SETS ) ... */

                /* Is there any cache set full ??? */
            for (s = 0; s < CACHE_SETS; s++)
                if (nrd_blocks_in_cache[s] >= ((CACHE_SIZE_BLOCKS * (num_nrd_servers-1)) / 2) )
                    break;

                /* The cache set [s] full, we have to send it over ... */
            if ( s < CACHE_SETS ) {
                dprintf("********** CACHE SET %d FULL : Send it over ...\n", s);

                /* Make the proper block mappings :
                 * Find an empty space on the first server
                 * (on the second if he's dead). All the blocks
                 * on the other servers will (and must) be the same ... */

                for (j = 0; j < CACHE_SIZE_BLOCKS; j+=2) {
                    int servblock;

                    servblock = -1;
                    for (i = 0; i < num_nrd_servers; i++) {

                        if (nrd_servers[i].alive == 0)
                            goto re_map_next_block;

                        /* If the new blocks are overwriting old ones
                         * invalidate the old blocks ... */
                        if ((i != num_nrd_servers-1) &&
                            (nrd_block_map[ nrd_cache[s][i].v_block_index[j] ].server != -1)) {
                            
                            srv = nrd_block_map[ nrd_cache[s][i].v_block_index[j] ].server;
                            blk   = nrd_block_map[ nrd_cache[s][i].v_block_index[j] ].block_no;
                            if ((srv >= 0) && (srv < num_nrd_servers-1) && (blk >= 0)) {
                                nrd_server_map[srv][blk] = NRD_PBLOCK_INVALID;
                                nrd_server_map[srv][blk+1] = NRD_PBLOCK_INVALID;

                                    /* Can I clean all the parity group ??? */
                                for (p = 0; p < num_nrd_servers-1; p++) {
                                    if ((nrd_server_map[p][blk] != NRD_PBLOCK_INVALID) ||
                                        (nrd_server_map[p][blk+1] != NRD_PBLOCK_INVALID))
                                        break;
                                }
                                    /* Yes, I can ... */
                                if (p >= num_nrd_servers-1) {
                                    dprintf("Cleaning Parity Group\n");
                                    for (p = 0; p < num_nrd_servers; p++) {
                                        nrd_server_map[p][blk] = NRD_PBLOCK_EMPTY;
                                        nrd_server_map[p][blk+1] = NRD_PBLOCK_EMPTY;
                                    }
                                }
                            }
                        }
                        srv = -1;
                        blk = -1;

                            /* Haven't found a physical block yet :
                             * Find an empty block on the new server ... */
                        if (servblock < 0) {

                            for (k = 0; k < NRD_HOST_BLOCKS; k+=2)
                                if ((nrd_server_map[i][k] == NRD_PBLOCK_EMPTY) &&
                                        (nrd_server_map[i][k+1] == NRD_PBLOCK_EMPTY))
                                    break;

                                /* Server is 95% full. Do some redundancy cleaning ... */
                            if ( k > (int)(0.95 * NRD_HOST_BLOCKS) ) {

                                fprintf( stderr, "Disk is almost full. Cleaning ...\n");
                                fflush( stderr );

                                /* Call the nrd_cleaner ... */
                                if ( ! nrd_cleaner() ) {
                                    fprintf( stderr, "Cleaner Error ...\n");
                                    fflush( stderr );
                                    exit(1);
                                }

                                    /* Try now ... */
                                for (k = 0; k < NRD_HOST_BLOCKS; k+=2)
                                    if ((nrd_server_map[i][k] == NRD_PBLOCK_EMPTY) &&
                                            (nrd_server_map[i][k+1] == NRD_PBLOCK_EMPTY))
                                        break;
                                if ( k < (int)(0.95 * NRD_HOST_BLOCKS) )
                                    servblock = k;
                                else {
                                    fprintf( stderr, "Cleaner cleaned nothing !!!\n");
                                    fflush( stderr );
                                    goto end_write_req;
                                }

                            } else
                                servblock = k;
                        } /* if (servblock == -1) ... */

                            /* Check if blocks are already full :
                             * this should not happen !! */
                        if ((nrd_server_map[i][servblock] != NRD_PBLOCK_EMPTY) ||
                                (nrd_server_map[i][servblock+1] != NRD_PBLOCK_EMPTY) ) {
                            fprintf( stderr, "Wrong Free Blocks Found ! Blocks must be symetrical !\n");
                            fprintf( stderr, "nrd_server_map[%d][%d]=%d - nrd_server_map[%d][%d]=%d\n",
                                        i, servblock, nrd_server_map[i][servblock],
                                        i, servblock+1, nrd_server_map[i][servblock+1]);
                            fflush( stderr );
                        }


                        /* Mark them as full ... */
                        nrd_server_map[i][servblock] = NRD_PBLOCK_FULL;
                        nrd_server_map[i][servblock+1] = NRD_PBLOCK_FULL;

                        /* Map the physical indexes on the block_index */
                        nrd_cache[s][i].p_block_index[j] = servblock;
                        nrd_cache[s][i].p_block_index[j+1] = servblock+1;

                            /* Is this the parity buffer ??? */
                        if ( i == num_nrd_servers-1 ) {

                                /* Zero the parity buffer ... */
                            bzero( &(nrd_cache[s][i].block_data[0]), CACHE_SIZE_BYTES );

                                /* Now calculate the parity into the buffer ... */
                            for (l = 0; l < NRD_BLOCK_SIZE; l++)
                                for (p = 0; p < num_nrd_servers-1; p++)
                                    if (nrd_servers[p].alive != 0) {
                                        nrd_cache[s][i].block_data[(j*SYS_BLOCK_SIZE)+l] ^=
                                            nrd_cache[s][p].block_data[(j*SYS_BLOCK_SIZE)+l];
                                    }
                        } else {
                                /* Update the Block Map ... */
                            nrd_block_map[ nrd_cache[s][i].v_block_index[j] ].server = i;
                            nrd_block_map[ nrd_cache[s][i].v_block_index[j] ].block_no = servblock;
                            nrd_block_map[ nrd_cache[s][i].v_block_index[j+1] ].server = i;
                            nrd_block_map[ nrd_cache[s][i].v_block_index[j+1] ].block_no = servblock+1;
                        }

                    }
re_map_next_block:

                }


                /* Prepare the cached write req header ... */
                req.type = htonl( NRD_REQ_RLB_CACHE_WRITE );
                req.blocknum = htonl( 0 );
                req.stblk = htonl( 0 );

                /* Send cache to all the servers ... */
                for (k = 0; k < num_nrd_servers; k++) {

                    /* Send the header ... */
                    dprintf("********** SENDING HEADER WRITE_CACHE ...\n");

                    /* If server is dead, try to reconnect
                     * to the crashed server ... */
                    if (nrd_servers[k].alive == 0) {

                        fprintf( stderr, "One server is DEAD !\n");
                        fflush( stderr );
#ifdef NET_TESTING
                        if ((remotehost = gethostbyname(nrd_servers[k].hostname)) == NULL) {
                            perror("gethostbyname ");
                            return(-1);
                        }

                        bzero((char *) &remoteaddr, sizeof(struct sockaddr_in));
                        bcopy(remotehost->h_addr,(char *)&remoteaddr.sin_addr,remotehost->h_length);
                        remoteaddr.sin_port     =   htons(NRD_SERVER_PORT);
                        remoteaddr.sin_family   =   remotehost->h_addrtype;
    
                        /*
                        ** Creating the socket ... */
                        if((nrd_servers[k].sockfd = socket(remotehost->h_addrtype, SOCK_STREAM, 0)) < 0) {
                            perror("socket_create ");
                            exit(1);
                        }

                        /*
                        ** Setting the non-blocking socket options ... */
                        if(setsockopt(nrd_servers[k].sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&j,
                                    sizeof(int)) == -1) {
                            perror("Seting TCP_NODELAY ");
                            exit(1);
                        }

                        /*
                        ** Connecting the socket ... */
                        if (connect(nrd_servers[k].sockfd, (struct sockaddr *)&remoteaddr,
                                    sizeof( struct sockaddr )) == 0 )
                            restore_nrd_server( k );
                        else 
                            continue;

                            /* Hangs machine if no space available ... */
#ifdef SET_TCP_BUFFER
                        j = NRD_SOCKBUFSIZE;
                        if(setsockopt(nrd_servers[k].sockfd, SOL_SOCKET, SO_SNDBUF, (char *)&j,
                                    sizeof(int)) == -1) {
                                perror("Seting SND_BUF failed ");
                                exit(1);
                            }
                        j = NRD_SOCKBUFSIZE;
                        if(setsockopt(nrd_servers[k].sockfd, SOL_SOCKET, SO_RCVBUF, (char *)&j,
                                    sizeof(int)) == -1) {
                                perror("Seting RCV_BUF failed ");
                                exit(1);
                        }
#endif
#endif
                    }


#ifdef NET_TESTING
                    if (send_tcp( nrd_servers[k].sockfd, &req, (3*sizeof(int)) ) == 0) {
                        fprintf( stderr, "Server %d not responding ...\n", k);
                        fflush(stderr);
                        if ((nrd_down_server != -1) && (nrd_down_server != k)) {
                            fprintf( stderr, "More than one servers crashed !\n");
                            fprintf( stderr, "Data is lost. Exiting ...\n");
                            exit(123);
                        } else
                            nrd_down_server = k;
                        nrd_servers[k].alive = 0;
                        close(nrd_servers[k].sockfd);

                        if ( k != num_nrd_servers-1 ) {
                                /* Now remove the blocks of the dead server
                                * from the parity ... */
                            for (l = 0; l < CACHE_SIZE_BYTES; l++)
                                nrd_cache[s][num_nrd_servers-1].block_data[l] ^=
                                nrd_cache[s][k].block_data[l];
                        }
                        
                        continue;
                    }
#endif

                    /* Transform the block indexes and the timestamp
                     * into network byte order */
                    for (i = 0;i < CACHE_SIZE_BLOCKS;i++) {
                        nrd_cache[s][k].v_block_index[i] = htonl( nrd_cache[s][k].v_block_index[i] );
                        nrd_cache[s][k].p_block_index[i] = htonl( nrd_cache[s][k].p_block_index[i] );
                    }
                    nrd_cache[s][k].timestamp = htonl( nrd_cache_timestamp );

                    /* Write the data to all the servers ... */
                    dprintf("********** SENDING CACHE DATA ...\n");
#ifdef NET_TESTING
                    if (send_tcp( nrd_servers[k].sockfd, (char *) &(nrd_cache[s][k]),
                                      sizeof(Nrd_Reliable_Cache)) == 0) {
                        fprintf( stderr, "Server %d not responding ...\n", k);
                        fflush(stderr);
                        if ((nrd_down_server != -1) && (nrd_down_server != k)) {
                            fprintf( stderr, "More than one servers crashed !\n");
                            fprintf( stderr, "Data is lost. Exiting ...\n");
                            exit(124);
                        } else
                            nrd_down_server = k;
                        nrd_servers[k].alive = 0;
                        close(nrd_servers[k].sockfd);

                        if ( k != num_nrd_servers-1 ) {
                                /* Now remove the blocks of the dead server
                                * from the parity ... */
                            for (l = 0; l < CACHE_SIZE_BYTES; l++)
                                nrd_cache[s][num_nrd_servers-1].block_data[l] ^=
                                nrd_cache[s][k].block_data[l];
                        }
                        
                        continue;
                    }
#endif

                } /* for all the servers ... */
                
                /* Increase the cache timestamp counter ... */
                nrd_cache_timestamp++;

                /* Increase the cache timeout counter if this is
                 * the victim set ... */
                if ( s == CACHE_SETS-1 ) {
                    for (i = 0; i < CACHE_SETS; i++)
                        if ( i != s )
                            nrd_cache_timeout[i]++;
                } else
                    nrd_cache_timeout[s] = 0;
                    

                /* Empty the Cache of the servers that
                 * are OK ... */
                dprintf("********** EMPTY THE CACHE ...\n");

                nrd_blocks_in_cache[s] = 0;
                for (i = 0; i < num_nrd_servers; i++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        if (nrd_servers[i].alive != 0) {
                            nrd_cache[s][i].v_block_index[j] = -1;
                            nrd_cache[s][i].p_block_index[j] = -1;
                        }

                /* If one server just crashed, move its blocks into
                 * another buffer ... */
                if (nrd_down_server != -1) {
                    for (i = 0; i < num_nrd_servers-1; i++)
                        if (nrd_servers[i].alive != 0)
                            break;

                        /* Move the data into another empty buffer ... */
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        nrd_cache[s][i].v_block_index[j] =
                            ntohl(nrd_cache[nrd_down_server][nrd_down_server].v_block_index[j]);
                    bcopy( nrd_cache[s][nrd_down_server].block_data,
                           nrd_cache[s][i].block_data, CACHE_SIZE_BYTES);

                        /* Empty the caches of all the buffers that
                         * have the crashed server ... */
                    for (s = 0; s < CACHE_SETS; s++)
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++) {
                        nrd_cache[s][nrd_down_server].v_block_index[j] = -1;
                        nrd_cache[s][nrd_down_server].p_block_index[j] = -1;
                    }
                }

            } /* If a set is full ... */

#else
            "YOU MUST EDIT nrd.h TO CONFIGURE the NETWORK RAMDISK"
#endif


#ifdef NRD_RELIABLE_PARITY
end_write_req:
#endif

        } /* End of Write request ... */

        else 
            fprintf( stderr, " Unknown Request. Discarded\n");

    }  /* while  ... */


    return 0;   /* pointless ... */
    
}



    /*
     ***************************************************
     *   Function: send_tcp()
     *   Arguments: 1) Socket file descriptor.
     *              2) Data buffer
     *              3) Size of the data.
     *   Description: Writing data to a TCP socket.
     ***************************************************
     */
int
send_tcp( int fd, char *buffer, int size ) {

    int i,j;

    i = 0;
    while(i < size ) {
/*      if((j = write( fd, (char *) (buffer + i), size - i) ) == -1) {*/
        if((j = send( fd, (char *) (buffer + i), size - i, 0)) == -1) {
            perror("Writing TCP Data");
            return 0;
        }
    i += j;
    }
    return 1;
}



    /*
     ***************************************************
     *   Function: recv_tcp()
     *   Arguments: 1) Socket file descriptor.
     *              2) Data buffer
     *              3) Size of the data.
     *   Description: Reading data from a TCP socket.
     ***************************************************
     */
int
recv_tcp( int fd, char *buffer, int size ) {

    int i,j;

    i = 0;
    while(i < size ) {
/*      if((j = read( fd, (char *) (buffer + i), size - i) ) == -1) { */
        if((j = recv( fd, (char *) (buffer + i), size - i, 0)) == -1) {
            perror("Reading TCP Data");
            return 0;
        }
    i += j;
    }
    return 1;
}



#ifdef NRD_RELIABLE_PARITY

    /*
     **************************************************************
     *   Function: nrd_cleaner()
     *   Arguments: None
     *   Description: Redundancy Cleaning Function ...
     **************************************************************
     */
int
nrd_cleaner ( void ) {

    int b, p, bcount = 0, ccount = 0, level = 0, l;
    int *srv, i, j = 0, clean_cache_blocks = 0, k;
    Nrd_BRead blockdata;
    Nrd_Reliable_Cache *clean_cache = NULL;
    IO_Request  creq;


        /* Redundance nrd_cleaner stage "level" :
         * Find the parity groups that have "level"
         * full blocks and clean them ... */
    srv = (int *) nrd_malloc( (num_nrd_servers-2) * sizeof(int) );
    clean_cache = (Nrd_Reliable_Cache *) nrd_malloc( num_nrd_servers * sizeof(Nrd_Reliable_Cache) );

    /* Empty the clean_cache ... */
    for (i = 0; i < num_nrd_servers; i++)
        for (j = 0; j < CACHE_SIZE_BLOCKS; j++) {
            clean_cache[i].p_block_index[j] = -1;
            clean_cache[i].v_block_index[j] = -1;
        }
        for (k = 0; k < CACHE_SIZE_BYTES; k++)
            clean_cache[i].block_data[j] = 0;
    clean_cache_blocks = 0;

    /* Start cleaning from the lower level (less full blocks
     * per parity group ... */
    for (level = 1; level < num_nrd_servers-1; level++) {

        for (b = 0; b < NRD_HOST_BLOCKS; b+=2) {

                /* Count how many blocks are full */
            bcount = 0;
            for (p = 0; p < num_nrd_servers-1; p++)
                if (nrd_server_map[p][b] == NRD_PBLOCK_FULL) {
                    srv[bcount] = p;
                    bcount++;
                }

                /* "level" blocks full. Clean it ... */
            if (bcount == level) {

                for (p = 0; p < bcount; p++) {

                    /* If one server is down, do not clean ... */
                    if (nrd_servers[srv[p]].alive == 0) {
                        fprintf( stderr, "Server %d not responding ...\n", srv[p]);
                        fprintf( stderr, "Not cleaning crashed server's blocks ...\n" );
                        fflush( stderr );
                        goto cleaning_next_block;
                    }

                    /* Reading the Block. Prepare the header ... */
                    creq.type = htonl( NRD_REQ_BLOCK_READV );
                    creq.blocknum = htonl( 2 );
                    creq.stblk = htonl( b );

#ifdef NET_TESTING
                    /* Get it from there ... */
                        /* Send the read request header ... */
                    if (send_tcp( nrd_servers[srv[p]].sockfd, (char *) &creq, (3*sizeof(int)) ) == 0) {
                        fprintf( stderr, "Server %d not responding ...\n", srv[p]);
                        if ((nrd_down_server != -1) && (nrd_down_server != srv[p])) {
                            fprintf( stderr, "More than one servers crashed !\n");
                            fprintf( stderr, "Data is lost. Exiting ...\n");
                            exit(125);
                        } else
                            nrd_down_server = srv[p];
                        nrd_servers[srv[p]].alive = 0;
                        close(nrd_servers[srv[p]].sockfd);
                    }

                    /* Read the data */
                    if ( recv_tcp( nrd_servers[srv[p]].sockfd, (char *) &blockdata,
                                    sizeof(Nrd_BRead)) == 0 ) {
                        fprintf( stderr, "Server %d not responding ...\n", srv[p]);
                        if ((nrd_down_server != -1) && (nrd_down_server != srv[p])) {
                            fprintf( stderr, "More than one servers crashed !\n");
                            fprintf( stderr, "Data is lost. Exiting ...\n");
                            exit(126);
                        } else
                            nrd_down_server = srv[p];
                        nrd_servers[srv[p]].alive = 0;
                        close(nrd_servers[srv[p]].sockfd);
                    }
#endif
                    /* If one server is down, do not clean ... */
                    if (nrd_servers[srv[p]].alive == 0) {
                        fprintf( stderr, "Server %d not responding ...\n", srv[p]);
                        fprintf( stderr, "Not cleaning crashed server's blocks ...\n" );
                        fflush( stderr );
                        goto cleaning_next_block;
                    }

                        /* Is the clean_cache full ??? */
                    if ( clean_cache_blocks < ((CACHE_SIZE_BLOCKS * (num_nrd_servers-1)) / 2) ) {

                        /* Put the blocks in the cache buffer ... */
                        for (i = 0; i < num_nrd_servers-1; i++) {
                            if (nrd_servers[i].alive != 0)
                                for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                                    if (clean_cache[i].v_block_index[j] == -1) {
                                        int flag = 1;
                                        for(k = 0; k < 2; k++)
                                            if ((j+k < CACHE_SIZE_BLOCKS) &&
                                                    (clean_cache[i].v_block_index[j+k] != -1))
                                                flag = 0;
                                        if (flag)   goto clean_findout;
                                    }
                        }
clean_findout:

                        /* Write the blocks there ... */
                        for (k = 0; k < 2; k++)
                            clean_cache[i].v_block_index[j+k] = blockdata.v_block_index[k];
                        bcopy( (char *)(blockdata.block_data),
                               (char *) &(clean_cache[i].block_data[j * SYS_BLOCK_SIZE]), NRD_BLOCK_SIZE );
                        clean_cache_blocks++;

                        /* The cache is full. Send it over ... */
                    } else {

                        /* Counting the full caches ... */
                        ccount ++;

                        /* Make the proper block mappings :
                        * Find an empty space on the first server
                        * All the blocks on the other servers will
                        * (and must) be the same ... */

                        for (j = 0; j < CACHE_SIZE_BLOCKS; j+=2) {
                            int servblock;

                            servblock = -1;
                            for (i = 0; i < num_nrd_servers; i++) {

                            /* If server is down, continue ... */
                            if (nrd_servers[i].alive == 0)
                                continue;

                            /* If the new blocks are overwriting old ones
                            * invalidate the old blocks ... */
                            if ((i != num_nrd_servers-1) &&
                                (clean_cache[i].v_block_index[j] >= 0)) {
                                int bl,serv;
                        
                                dprintf("Invalidating old blocks\n");
                                serv = nrd_block_map[ clean_cache[i].v_block_index[j] ].server;
                                bl   = nrd_block_map[ clean_cache[i].v_block_index[j] ].block_no;

                                if ((serv >= 0) && (serv < num_nrd_servers-1) && (bl >= 0)) {
                                    nrd_server_map[serv][bl] = NRD_PBLOCK_INVALID;
                                    nrd_server_map[serv][bl+1] = NRD_PBLOCK_INVALID;

                                        /* Can I clean all the parity group ??? */
                                    for (p = 0; p < num_nrd_servers-1; p++) {
                                        if ((nrd_server_map[p][bl] != NRD_PBLOCK_INVALID) ||
                                            (nrd_server_map[p][bl+1] != NRD_PBLOCK_INVALID))
                                            break;
                                    }
                                        /* Yes, I can ... */
                                    if (p >= num_nrd_servers-1) {
                                        dprintf("Cleaning Parity Group\n");
                                        for (p = 0; p < num_nrd_servers; p++) {
                                            nrd_server_map[p][bl] = NRD_PBLOCK_EMPTY;
                                            nrd_server_map[p][bl+1] = NRD_PBLOCK_EMPTY;
                                        }
                                    }
                                }
                            }

                                /* Haven't found a physical block yet :
                                * Find an empty block on the new server ... */
                            if (servblock == -1) {
                                for (k = 0; k < NRD_HOST_BLOCKS; k+=2)
                                    if ((nrd_server_map[i][k] == NRD_PBLOCK_EMPTY) &&
                                            (nrd_server_map[i][k+1] == NRD_PBLOCK_EMPTY))
                                        break;

                                    /* Server is full. Exit the nrd_cleaner ... */
                                if ( k >= NRD_HOST_BLOCKS ) {

                                    fprintf( stderr, "Servers are FULL !\n");
                                    free(clean_cache);
                                    free(srv);
                                    return 1;
                                } else
                                    servblock = k;
                            }

                                /* Check if blocks are already full :
                                * this should not happen !! */
                            if ((nrd_server_map[i][servblock] != NRD_PBLOCK_EMPTY) ||
                                    (nrd_server_map[i][servblock+1] != NRD_PBLOCK_EMPTY) ) {
                                fprintf( stderr, "Wrong Free Blocks Found ! Blocks must be symetrical !\n");
                                fprintf( stderr, "nrd_server_map[%d][%d]=%d - nrd_server_map[%d][%d]=%d\n",
                                            i, servblock, nrd_server_map[i][servblock],
                                            i, servblock+1, nrd_server_map[i][servblock+1]);
                            }

                            /* Mark them as full ... */
                            nrd_server_map[i][servblock] = NRD_PBLOCK_FULL;
                            nrd_server_map[i][servblock+1] = NRD_PBLOCK_FULL;

                            /* Map the physical indexes on the block_index */
                            clean_cache[i].p_block_index[j] = servblock;
                            clean_cache[i].p_block_index[j+1] = servblock+1;

                                /* Is this the parity buffer ??? */
                            if ( i == num_nrd_servers-1 ) {
                                    /* Zero the parity buffer ... */
                                bzero( clean_cache[i].block_data, CACHE_SIZE_BYTES );

                                    /* Now calculate the parity into the buffer ... */
                                for (l = 0; l < NRD_BLOCK_SIZE; l++)
                                    for (p = 0; p < num_nrd_servers-1; p++)
                                        if (nrd_servers[p].alive != 0)
                                            clean_cache[i].block_data[(j*SYS_BLOCK_SIZE)+l] ^=
                                                clean_cache[p].block_data[(j*SYS_BLOCK_SIZE)+l];
                            } else {
                                    /* Update the Block Map ... */
                                nrd_block_map[ clean_cache[i].v_block_index[j] ].server = i;
                                nrd_block_map[ clean_cache[i].v_block_index[j] ].block_no = servblock;
                                nrd_block_map[ clean_cache[i].v_block_index[j+1] ].server = i;
                                nrd_block_map[ clean_cache[i].v_block_index[j+1] ].block_no = servblock+1;
                            }

                            } /* for (i = 0; i < num_nrd_servers; i++) */

                        } /* for (j = 0; j < CACHE_SIZE_BLOCKS ... */


                        /* Prepare the cached write creq header ... */
                        creq.type = htonl( NRD_REQ_RLB_CACHE_WRITE );
                        creq.blocknum = htonl( 0 );
                        creq.stblk = htonl( 0 );

#ifdef NET_TESTING
                        /* Send cache to all the servers ... */
                        for (k = 0; k < num_nrd_servers; k++) {

                            /* If server is dead ignore it ... */
                            if (nrd_servers[k].alive == 0)
                                continue;

                            /* Transform the block indexes and the timestamp
                            * into network byte order */
                            for (i = 0;i < CACHE_SIZE_BLOCKS;i++) {
                                clean_cache[k].v_block_index[i] = htonl( clean_cache[k].v_block_index[i] );
                                clean_cache[k].p_block_index[i] = htonl( clean_cache[k].p_block_index[i] );
                            }
                            clean_cache[k].timestamp = htonl( nrd_cache_timestamp );

                            /* Send the header ... */
                            if (send_tcp( nrd_servers[k].sockfd, (char *) &creq, (3*sizeof(int)) ) == 0) {
                                fprintf( stderr, "Server %d not responding ...\n", k);
                                if ((nrd_down_server != -1) && (nrd_down_server != k)) {
                                    fprintf( stderr, "More than one servers crashed !\n");
                                    fprintf( stderr, "Data is lost. Exiting ...\n");
                                    exit(126);
                                } else
                                    nrd_down_server = k;
                                nrd_servers[k].alive = 0;
                                close(nrd_servers[k].sockfd);
                                continue;
                            }

                            /* Write the data to all the servers ... */
                            if (send_tcp( nrd_servers[k].sockfd, (char *) &(clean_cache[k]),
                                            sizeof(Nrd_Reliable_Cache)) == 0) {
                                fprintf( stderr, "Server %d not responding ...\n", k);
                                if ((nrd_down_server != -1) && (nrd_down_server != k)) {
                                    fprintf( stderr, "More than one servers crashed !\n");
                                    fprintf( stderr, "Data is lost. Exiting ...\n");
                                    exit(127);
                                } else
                                    nrd_down_server = k;
                                nrd_servers[k].alive = 0;
                                close(nrd_servers[k].sockfd);
                                continue;
                            }

                        } /* for all the servers ... */
#endif
                
                        /* Increase the cache timestamp counter ... */
                        nrd_cache_timestamp++;

                        /* Empty the clean_cache ... */
                        clean_cache_blocks = 0;
                        for (i = 0; i < num_nrd_servers; i++)
                            for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                                if (nrd_servers[i].alive != 0) {
                                    clean_cache[i].v_block_index[j] = -1;
                                    clean_cache[i].p_block_index[j] = -1;
                                }

                        /* Write the blocks in the empty cache ... */
                        for (k = 0; k < 2; k++)
                            clean_cache[0].v_block_index[k] = blockdata.v_block_index[k];
                        bcopy( (char *)(blockdata.block_data), (char *) &(clean_cache[0].block_data[0]),
                                    NRD_BLOCK_SIZE );
                        clean_cache_blocks++;
                    }

                    /* Check the cleaning threshold */
                    if (ccount > NRD_MAX_CLEANING) {
                        dprintf(stdout, "Cleaned %d caches.\n", ccount);

                        free(clean_cache);
                        free( srv );
                        return 1;
                    }

                } /* for (bcount ... */

            }   /* if (bcount == lev ... */

cleaning_next_block:

        } /* for (b = 0 ... */

    }   /* for (lev ... */

    free(clean_cache);
    free( srv );

    return 1;
}


    /*
     **************************************************************
     *   Function: restore_nrd_server()
     *   Arguments: The server number to restore ...
     *   Description: Crashed Server Restoration function.
     *                The restoration cost is quite high, but
     *                generally we do not care, because it's not
     *                supposed to happen often ... 
     **************************************************************
     */
void
restore_nrd_server ( int rsrv ) {
    
    int i, j, k, b, rest_cache_blocks = 0;
    Nrd_Reliable_Cache rest_cache;
    IO_Request  creq;
    char *tmpdata, *restdata;

    fprintf( stdout, "Server %d is back on-line ...\n", rsrv);
    fprintf( stdout, "Restoring its blocks from parity. Wait ...\n");

        /* Initialize the structs ... */
    tmpdata = (char *) malloc( NRD_BLOCK_SIZE );
    restdata = (char *) malloc( NRD_BLOCK_SIZE );
    for (j = 0; j < CACHE_SIZE_BLOCKS; j++) {
        rest_cache.v_block_index[j] = -1;
        rest_cache.p_block_index[j] = -1;
    }

    for (b = 0; b < NRD_HOST_BLOCKS; b+=2) {

        /* Find out if the block is full (or invalid) */
        if (nrd_server_map[rsrv][b] != NRD_PBLOCK_EMPTY) {

            /* Prepare the header ... */
            creq.type = htonl( NRD_REQ_BLOCK_READ );
            creq.blocknum = htonl( NRD_BLOCK_SIZE/SYS_BLOCK_SIZE );
            creq.stblk = htonl( b );

            bzero( restdata, NRD_BLOCK_SIZE );
            for (i = 0; i < num_nrd_servers; i++)
                if (i != rsrv) {
    
#ifdef NET_TESTING
                /* Get it from there ... */
                    /* Send the read request header ... */
                if (send_tcp( nrd_servers[i].sockfd, (char *) &creq,
                                        (3*sizeof(int)) ) == 0) {
                    fprintf( stderr, "Server crashed on restoration !\n");
                    fprintf( stderr, "Cannot continue ...\n");
                    exit( 201 );
                }

                    /* Read the data */
                if ( recv_tcp( nrd_servers[i].sockfd, (char *) tmpdata,
                                        NRD_BLOCK_SIZE) == 0 ) {
                    fprintf( stderr, "Server crashed on restoration !\n");
                    fprintf( stderr, "Cannot continue ...\n");
                    exit( 202 );
                }
#endif
                    /* Calculate the block from the parity ... */
                for (k = 0; k < NRD_BLOCK_SIZE; k++)
                    restdata[k] ^= tmpdata[k];
                }

                /* Put the restored block in the cache ... */
                if ( rest_cache_blocks < (CACHE_SIZE_BLOCKS/2) ) {

                    /* Put the blocks in the cache buffer ... */
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        if (rest_cache.v_block_index[j] == -1) {
                            int flag = 1;
                            for(k = 0; k < 2; k++)
                                if ((j+k < CACHE_SIZE_BLOCKS) &&
                                        (rest_cache.v_block_index[j+k] != -1))
                                    flag = 0;
                            if (flag)   goto rest_findout;
                        }
rest_findout:

                    /* Write the blocks there ... */
                    if (nrd_server_map[rsrv][b] == NRD_PBLOCK_FULL) {

                            /* Find the virtual block number from the nrd_block_map ...
                             * This costs a lot, but generally we do not
                             * care about the restoration cost ... */
                        for (i = 0; i < nrd_size_blocks; i++)
                            if (( nrd_block_map[i].server == rsrv ) &&
                                ( nrd_block_map[i].block_no == b ))
                                break;

                            /* Write the virtual block numbers found ... */
                        if (i >= nrd_size_blocks ) {
/*                          fprintf( stderr, "Virtual block not found\n");*/
                            i = -1;
                        }
                        for (k = 0; k < 2; k++)
                            rest_cache.v_block_index[j+k] = i + k;
                    }
                    /* Map the physical indexes on the block_index */
                    for (k = 0; k < 2; k++)
                        rest_cache.p_block_index[j+k] = b + k;
                    bcopy( (char *)(restdata),
                           (char *) &(rest_cache.block_data[j * SYS_BLOCK_SIZE]), NRD_BLOCK_SIZE );
                    rest_cache_blocks++;

                    /* The cache is full. Send it over ... */
                } else {

                    /* Prepare the cached write creq header ... */
                    creq.type = htonl( NRD_REQ_RLB_CACHE_WRITE );
                    creq.blocknum = htonl( 0 );
                    creq.stblk = htonl( 0 );
#ifdef NET_TESTING

                    /* Transform the block indexes and the timestamp
                    * into network byte order */
                    for (i = 0;i < CACHE_SIZE_BLOCKS;i++) {
                        rest_cache.v_block_index[i] = htonl( rest_cache.v_block_index[i] );
                        rest_cache.p_block_index[i] = htonl( rest_cache.p_block_index[i] );
                    }
                    rest_cache.timestamp = htonl( nrd_cache_timestamp );

                    /* Send the header ... */
                    if (send_tcp( nrd_servers[rsrv].sockfd, (char *) &creq, (3*sizeof(int)) ) == 0) {
                        fprintf( stderr, "Server crashed on restoration !\n");
                        fprintf( stderr, "Cannot continue ...\n");
                        exit( 203 );
                    }

                    /* Write the data to the restored server ... */
                    if (send_tcp( nrd_servers[rsrv].sockfd, (char *) &rest_cache,
                                    sizeof(Nrd_Reliable_Cache)) == 0) {
                        fprintf( stderr, "Server crashed on restoration !\n");
                        fprintf( stderr, "Cannot continue ...\n");
                        exit( 203 );
                    }
#endif
                
                    /* Increase the cache timestamp counter ... */
                    nrd_cache_timestamp++;

                    /* Empty the rest_cache ... */
                    rest_cache_blocks = 0;
                    for (j = 0; j < CACHE_SIZE_BLOCKS; j++)
                        rest_cache.v_block_index[j] = -1;
                        rest_cache.p_block_index[j] = -1;
                    
                    /* Write the blocks in the empty cache ... */
                    if (nrd_server_map[rsrv][b] == NRD_PBLOCK_FULL) {

                            /* Find the virtual block number from the nrd_block_map ...
                             * This costs a lot, but generally we do not
                             * care about the restoration cost ... */
                        for (i = 0; i < nrd_size_blocks; i++)
                            if (( nrd_block_map[i].server == rsrv ) &&
                                ( nrd_block_map[i].block_no == b ))
                                break;

                            /* Write the virtual block numbers found ... */
                        if (i >= nrd_size_blocks ) {
/*                          fprintf( stderr, "Virtual block not found\n");*/
                            i = -1;
                        }
                        for (k = 0; k < 2; k++)
                            rest_cache.v_block_index[k] = i + k;
                    }
                    /* Map the physical indexes on the block_index */
                    for (k = 0; k < 2; k++)
                        rest_cache.p_block_index[k] = b + k;
                    bcopy( (char *)(restdata), (char *) rest_cache.block_data,
                                NRD_BLOCK_SIZE );
                    rest_cache_blocks++;
                }

        } /* if (nrd_server_map[rsrv][b] != NRD_PBLOCK_EMPTY)  ... */

        if ( b % 1000 == 0 ) {
            fprintf( stdout, ".");
            fflush( stdout );
        }
    } /* for (b = 0; b < NRD_HOST_BLOCKS; ... */
    
    free( tmpdata );
    free( restdata );
    nrd_down_server = -1;
    nrd_servers[rsrv].alive = 1;
    fprintf( stdout, "Restoration Finished.\n");

    return;
}
#endif



    /*
     **************************************************************
     *   Function: nrd_malloc()
     *   Arguments: Number of bytes to allocate.
     *   Description: Allocating memory, and returning error code
     *                in case of failure.
     **************************************************************
     */
void *
nrd_malloc ( int bytes ) {

    void * ptr = NULL;

    if ( bytes < 0 ) {
        fprintf( stderr, "Negative argument to nrd_malloc. Exiting ...\n");
        exit(1);
    }

    ptr = (void *) malloc( bytes );

    if ( ptr != NULL )
        return ptr;
    else {
        fprintf( stderr, "nrd_malloc %d bytes\n", bytes);
        perror("nrd : malloc");
        exit(1);
    }
}


    /*
     **************************************************************
     *   Function:    nrd_client_sighandler()
     *   Arguments:   Signal number
     *   Description: Responds to signals sent to server.
     **************************************************************
     */
void
nrd_client_sighandler( int signo ) {

    int i = 0;

    for (i = 0; i < num_nrd_servers; i++)
        if (close(nrd_servers[i].sockfd) == -1)
            perror( "Closing socket" );

    fprintf( stderr, "Caught SIGINT signal. Exiting ...\n" );

    exit( 0 );
}



    /*
     **************************************************************
                          This is the End ...
     **************************************************************
     */


