

/* ******************************************************************

    File:          nrd_server.c
    Author:        Mike Flouris (C)1997-1998

    Description:        
        User level server process handling remote block requests.
        It accepts data for block out and sends data for block in.
    Requires :
        The header file of Network Ramdisk device driver ... (nrd.h) 

   ******************************************************************
*/

#include <stdio.h>          /* I/O definitions */
#include <sys/param.h>      /* DEV_BSIZE definition */
#include <signal.h>         /* Signal definitions */
#include <string.h>         /* Bzero definition */
#include <fcntl.h>
#include <unistd.h>         /* Read/write definitions */
#include <errno.h>          /* Perror definition */
#include <sys/time.h>
#include <sys/types.h>      /* select, FDXX definition */
#include <sys/socket.h>     /* Socket definition */
#include <netinet/in.h>     /* Internet socket definition */
#include <netinet/tcp.h>    /* TCP_NODELAY definition */
#include <netdb.h>
#include <sys/stat.h>

#include "nrd.h"        /* Network RamDisk include file ... */

/* Produce trace dumps on DISK */
#undef TRACE_DUMP

#undef NRDDEBUG

#ifdef TRACE_DUMP
    FILE    *tdfile;
    char    *tdfilename;
#endif

#define NRD_SOCKBUFSIZE  256*1024   /* TCP buffer size ... */
            
char    nrddisk [NRD_HOST_BYTES];   /* Data repository. */

void    handle_request();
void    nrd_sighandler();
void    shutdown_server();

int blocks_sent = 0;
int bytes_sent  = 0;
int blocks_received = 0;
int bytes_received  = 0;


/*
===========================================================================
  Function:     main
  Arguments:    None
  Description:  
     Creates a socket, binds it to NRD_SERVER_PORT and calls
    handle_request for each connection. Every connection should be
    handled by a new process. Let just one connection for the moment.
===========================================================================
*/

int
main( void )
{
    int     sockfd, newsockfd;
    struct  sockaddr_in serveraddr;
    struct  sockaddr_in clientaddr;
    struct  hostent    *clienthost;
    int     clientaddrlen = sizeof(clientaddr);
    int     j = 1, i = 0;
    char    * buffer;



    /* NRD Signal Handler to perform necessary cleanups */
    signal(SIGINT, nrd_sighandler);

    /*
    ** Creating the socket ... */
    if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        perror("NRD : socket_create");
        exit(1);
    }

    /*
    ** Setting the socket options :
    ** TCP_NODELAY : this makes the socket I/O asynchronoys(non-blocking) but
                     causes much machine load because of busy waiting on recv.
                     This is why we must use select() ... */
    if(setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &j, sizeof(int)) == -1) {
        perror("NRD : Seting TCP_NODELAY");
        exit(1);
    }

    bzero((char *) &serveraddr, sizeof(struct sockaddr_in));
    serveraddr.sin_family       =   AF_INET;
    serveraddr.sin_addr.s_addr  =   htonl(INADDR_ANY);
    serveraddr.sin_port     =   htons(NRD_SERVER_PORT);
    
    /*
    ** Binding the socket ... */
    if(bind(sockfd, (struct sockaddr *) &serveraddr, 
            sizeof(struct sockaddr_in)) < 0 ) {
        extern  int errno;
        
        printf("Errno is %d\n", errno);
        perror("NRD : socket_bind");
        close(sockfd);
        exit(1);
    }

    /*
    ** Listening to the socket ... */
    printf("Waiting for connection ...\n");
    listen(sockfd, 5);
    
    while(1) {
        if((newsockfd = accept(sockfd, (struct sockaddr *) &clientaddr,
            &clientaddrlen)) < 0) {
            perror("NRD : socket_accept");
            close(sockfd);
            exit(1);
        }
#ifdef  SET_TCP_BUFFER          /* Hangs machine if no space available ... */
        j = NRD_SOCKBUFSIZE;
                if(setsockopt(newsockfd, SOL_SOCKET, SO_SNDBUF, (char *)&j, sizeof(int)) == -1)
                {
                        perror("NRD : Seting SND_BUF");
                        exit(1);
                }
                j = NRD_SOCKBUFSIZE;
                if(setsockopt(newsockfd, SOL_SOCKET, SO_RCVBUF, (char *)&j, sizeof(int)) == -1)
                {
                        perror("NRD : Seting RCV_BUF");
                        exit(1);
                }
#endif
        printf("Got a Call from 0x%x ...\n", clientaddr.sin_addr.s_addr);
        handle_request(newsockfd);
    }

    handle_request(sockfd);

    return 0;
    
}



/*
===========================================================================
  Function: nrd_sighandler
  Arguments:    Signal number
  Description:  Responds to signals sent to server.
===========================================================================
*/
void    nrd_sighandler(int signo)
{
    printf("Blocks sent %d, Blocks received %d\n", blocks_sent, blocks_received);
    printf("Bytes  sent %d, Bytes  received %d\n", bytes_sent, bytes_received);

#ifdef  TRACE_DUMP
    fclose( tdfile );
#endif
    exit( 0 );
}


/*
====================================================================
  Function: handle_request
  Arguments:    Socket file descriptor
  Description:  Handles a connection performing blockin/blockout 
        requests.
====================================================================
*/
void    handle_request(int fd)
{
    unsigned    request;        /* Type of request (read/write) */
    unsigned    num_blocks, num_bytes;  /* Number of requested blocks/bytes */
    unsigned    blkno, byteno;  /* Number of starting block / byte */
    IO_Request  transfer;       /* Header for the I/O Request */
    char        ack = 0;        /* Ack flag ... */
    int         i, j;
                                /* fd_sets for select() ... */
    fd_set      rd_fdset,wr_fdset;  
#ifdef NRD_UNRELIABLE_CACHE
    Nrd_Cache   nrd_cache_buf;
#elif defined(NRD_RELIABLE_PARITY)
    Nrd_BRead   rdbuf;
    Nrd_Reliable_Cache  nrd_cache_buf;
    int         nrd_host_map[ NRD_HOST_BLOCKS ];
#endif
#ifdef  NRD_INSTRUMENT
    struct  timeval tv1, tv2;
    struct  timezone tz;
#endif  /*  NRD_INSTRUMENT  */  

#if defined(NRD_RELIABLE_PARITY)
    for (i = 0; i < NRD_HOST_BLOCKS; i++)
        nrd_host_map[i] = -1;
#endif

#ifdef TRACE_DUMP
    /* Initialize Trace Dump file ... */
    tdfilename = (char *) malloc ( 32 * sizeof(char) );
    if ( gethostname( tdfilename, 26 ) == -1) {
        perror("NRD : gethostname");
        exit(1);
    }
    strcat( tdfilename, ".td");
    tdfile = fopen( tdfilename, "w");
#endif


    while (1) {     /* One handle_request per client ... */


#ifdef  NRD_INSTRUMENT
        gettimeofday(&tv1, &tz);
#endif  /*  NRD_INSTRUMENT  */

        /* Read the Request ... */
    recv_tcp( fd, (char *) &transfer, (3 * sizeof(int)) );

        /* Transform the header into host byte order */
    request = ntohl( transfer.type );
    num_blocks = ntohl( transfer.blocknum );
    num_bytes  = ntohl( transfer.blocknum );
    blkno   = ntohl( transfer.stblk );
    byteno  = ntohl( transfer.stblk );



#ifdef  NRD_INSTRUMENT
    gettimeofday(&tv2, &tz);
    printf("Time to read request in usec %d\n", 
           tv2.tv_usec - tv1.tv_usec + (tv2.tv_sec - tv1.tv_sec)*1000000);
#endif  /*  NRD_INSTRUMENT  */

            /* Satisfy the request ... */
    switch(request)
    {

/* ************************************ */
    case NRD_REQ_BLOCK_WRITE:
/* ************************************ */
        blocks_received += num_blocks;
#ifdef  NRDDEBUG
        printf("NRD Request    : WRITE\n");
        printf("Starting Block : %d\n", blkno );
        printf("Request Bytes  : %d\n", (num_blocks * SYS_BLOCK_SIZE) );
#endif

#ifdef TRACE_DUMP
        fprintf( tdfile, "write  %d  %d\n", num_blocks, blkno );
#endif

        if ( (blkno + num_blocks) > NRD_HOST_BLOCKS) {
            printf("******* OUT OF BOUNDS (req:%d, block:%d)...\n",
                request, (blkno + num_blocks) );
            break;
        }

            /* Reading  data from the socket ... */
        recv_tcp( fd, (char *) &(nrddisk[ blkno * SYS_BLOCK_SIZE ]), (num_blocks * SYS_BLOCK_SIZE));

#ifdef  NRDDEBUG
        printf("Transfer ended ...\n");
#endif  /*  NRDDEBUG  */            

#ifdef  NRD_INSTRUMENT
        gettimeofday(&tv1, &tz);
        printf("Time to send out %d blocks in usec %d\n",num_blocks, 
            tv1.tv_usec - tv2.tv_usec + (tv1.tv_sec - tv2.tv_sec)*1000000);
#endif  /*  NRD_INSTRUMENT  */
    break;
            

/* ************************************ */
    case NRD_REQ_BLOCK_READ:
/* ************************************ */
        blocks_sent += num_blocks;
#ifdef  NRDDEBUG
        printf("NRD Request    : READ\n");
        printf("Starting Block : %d\n", blkno );
        printf("Request Bytes  : %d\n", (num_blocks * SYS_BLOCK_SIZE) );
#endif

#ifdef TRACE_DUMP
        fprintf( tdfile, "read  %d  %d\n", num_blocks, blkno );
#endif

        if ( (blkno + num_blocks) > NRD_HOST_BLOCKS) {
            printf("******* OUT OF BOUNDS (req:%d, block:%d)...\n",
                request, (blkno + num_blocks) );
            break;
        }

            /* Write the data to the socket ... */
        send_tcp( fd, (char *) &(nrddisk[ blkno * SYS_BLOCK_SIZE ]), (num_blocks * SYS_BLOCK_SIZE));

#ifdef  NRDDEBUG
        printf("Transfer ended ...\n");
#endif  /*  NRDDEBUG  */            

#ifdef  NRD_INSTRUMENT
        gettimeofday(&tv1, &tz);
        printf("Time to send in %d blocks in usec %d\n",num_blocks, 
           tv1.tv_usec - tv2.tv_usec + (tv1.tv_sec - tv2.tv_sec)*1000000);
#endif  /*  NRD_INSTRUMENT  */
    break;

#ifdef NRD_UNRELIABLE_CACHE
/* ************************************ */
    case NRD_REQ_CACHE_WRITE :
/* ************************************ */

#ifdef  NRDDEBUG
        printf("NRD Request    : WRITE CACHE\n");
        printf("Cache Size     : %d bytes (%d NRD blocks)\n",
               CACHE_SIZE_BYTES, CACHE_SIZE_BLOCKS);
#endif

        /* Reading data from the socket ... */
        recv_tcp( fd, (char *) &nrd_cache_buf, sizeof(Nrd_Cache) );

        /* Transform the block indexes into host byte order */
        for (i = 0;i < CACHE_SIZE_BLOCKS;i++)
            nrd_cache_buf.block_index[i] = ntohl( nrd_cache_buf.block_index[i] );

        /* Write the data to the memory ... */
        for (i = 0; i < CACHE_SIZE_BLOCKS; i+=2) {
            if ( ( nrd_cache_buf.block_index[i] > NRD_HOST_BLOCKS) ||
                    ( nrd_cache_buf.block_index[i] < 0 ) )
                printf("******* OUT OF BOUNDS (Block:%d)\n", nrd_cache_buf.block_index[i]);
            else {
                bcopy( &(nrd_cache_buf.block_data[i * SYS_BLOCK_SIZE]),
                        &(nrddisk[ (nrd_cache_buf.block_index[i]) * SYS_BLOCK_SIZE ]),
                        NRD_BLOCK_SIZE /* = 2*SYS_BLOCK_SIZE */ );
                blocks_received+= 2;
                /*printf("Writing Block %d\n", nrd_cache_buf.block_index[i]);*/
            }
        }

#ifdef  NRDDEBUG
        printf("Cache Blocks   : ");
        for (i = 0; i < CACHE_SIZE_BLOCKS; i++)
            printf("%d ", nrd_cache_buf.block_index[i] );
        printf("\n");
#endif  /*  NRDDEBUG  */            

#ifdef  NRD_INSTRUMENT
        gettimeofday(&tv1, &tz);
        printf("Time to send out %d blocks in usec %d\n",num_blocks, 
            tv1.tv_usec - tv2.tv_usec + (tv1.tv_sec - tv2.tv_sec)*1000000);
#endif  /*  NRD_INSTRUMENT  */

    break;

#elif defined(NRD_RELIABLE_PARITY)
/* ************************************ */
    case NRD_REQ_BLOCK_READV :
/* ************************************ */
        blocks_sent += num_blocks;

        if ( (blkno + num_blocks) > NRD_HOST_BLOCKS) {
            printf("******* OUT OF BOUNDS (req:%d, block:%d)...\n",
                request, (blkno + num_blocks) );
            break;
        }

            /* Write the data to the socket ... */
        bzero( rdbuf, sizeof(Nrd_BRead) );
        bcopy( (char *) &(nrddisk[ blkno * SYS_BLOCK_SIZE ]), rdbuf.block_data,
               (num_blocks * SYS_BLOCK_SIZE));
        rdbuf.v_block_index[0] = nrd_host_map[ blkno ];
        rdbuf.v_block_index[1] = nrd_host_map[ blkno+1 ];
        send_tcp( fd, (char *) &rdbuf, sizeof(Nrd_BRead) );

    break;

/* ************************************ */
    case NRD_REQ_RLB_CACHE_WRITE :
/* ************************************ */

#ifdef  NRDDEBUG
        printf("NRD Request    : WRITE RELIABLE CACHE\n");
        printf("Cache Size     : %d bytes (%d NRD blocks)\n",
               CACHE_SIZE_BYTES, CACHE_SIZE_BLOCKS);
#endif

        /* Reading data from the socket ... */
        recv_tcp( fd, (char *) &nrd_cache_buf, sizeof(Nrd_Reliable_Cache) );

        /* Transform the block indexes into host byte order */
        for (i = 0;i < CACHE_SIZE_BLOCKS;i++) {
            nrd_cache_buf.p_block_index[i] = ntohl( nrd_cache_buf.p_block_index[i] );
            nrd_cache_buf.v_block_index[i] = ntohl( nrd_cache_buf.v_block_index[i] );
        }
        nrd_cache_buf.timestamp = ntohl( nrd_cache_buf.timestamp );

        /* Ignoring the timestamp for the moment ... */

        /* Write the data to the memory ... */
        for (i = 0; i < CACHE_SIZE_BLOCKS; i+=2) {
            if ( ( nrd_cache_buf.p_block_index[i] > NRD_HOST_BLOCKS) ||
                    ( nrd_cache_buf.p_block_index[i] < 0 ) )
                printf("******* OUT OF BOUNDS (Block:%d)\n", nrd_cache_buf.p_block_index[i]);
            else {
                    /* Write the data to memory ... */
                bcopy( &(nrd_cache_buf.block_data[i * SYS_BLOCK_SIZE]),
                        &(nrddisk[ (nrd_cache_buf.p_block_index[i]) * SYS_BLOCK_SIZE ]),
                        NRD_BLOCK_SIZE /* = 2*SYS_BLOCK_SIZE */ );

                    /* Update the host map with the virtual block numbers ... */
                nrd_host_map[ nrd_cache_buf.p_block_index[i] ]   = nrd_cache_buf.v_block_index[i];
                nrd_host_map[ nrd_cache_buf.p_block_index[i+1] ] = nrd_cache_buf.v_block_index[i+1];

                blocks_received+= 2;
/*              printf("Writing Block %d\n", nrd_cache_buf.p_block_index[i]);*/
            }
        }

#ifdef  NRDDEBUG
        printf("Virtual Cache Blocks   : ");
        for (i = 0; i < CACHE_SIZE_BLOCKS; i++)
            printf("%d ", nrd_cache_buf.v_block_index[i] );
        printf("\n");
        printf("Physical Cache Blocks   : ");
        for (i = 0; i < CACHE_SIZE_BLOCKS; i++)
            printf("%d ", nrd_cache_buf.p_block_index[i] );
        printf("\n");
#endif  /*  NRDDEBUG  */

#ifdef  NRD_INSTRUMENT
        gettimeofday(&tv1, &tz);
        printf("Time to send out %d blocks in usec %d\n",num_blocks, 
            tv1.tv_usec - tv2.tv_usec + (tv1.tv_sec - tv2.tv_sec)*1000000);
#endif  /*  NRD_INSTRUMENT  */

    break;
#endif
                                /* Asking server on Virtual Memory Statistics */
/* ************************************ */
        case NRD_REQ_STATS:
/* ************************************ */
        fprintf(stderr, "Sending VM stats NOT yet implemented ...\n");
    break;

                                /* Shutting down this server */
/* ************************************ */
        case NRD_REQ_STOP_SERVER:
/* ************************************ */
        printf("Shutting down ...\n");
        return;
    break;

/* ************************************ */
    default:
/* ************************************ */
        fprintf( stderr, "Invalid request %d|%d|%d\n", 
            (transfer.type),(transfer.blocknum),(transfer.stblk) );
    break;

    }

#ifdef  NRDDEBUG
/*  printf("Restarting loop ... \n");*/
#endif  /*  NRDDEBUG  */            
    }  /* while (1) ... */
    
    return ; /* pointless ;) ... */


} /* handle request ... */


/*
 * TCP write function.
 * Used to minimize the size of the code ... */
int
send_tcp( int fd, char *buffer, int size ) {

    int i,j;

    i = 0;
    while(i < size ) {
/*      if((j = write( fd, (char *) (buffer + i), size - i) ) == -1) {*/
        if((j = send( fd, (char *) (buffer + i), size - i, 0)) == -1) {
            perror("Writing TCP Data");
            return 0;
        }
    i += j;
    }
    return 1;
}

/*
 * TCP read function.
 * Used to minimize the size of the code ... */
int
recv_tcp( int fd, char *buffer, int size ) {

    int i,j;
    fd_set soset;

    FD_ZERO(&soset);
    FD_SET(fd, &soset);

    while ( 1 ) {
        /* Wait until data arrives ... */
        if (select( fd+1, &soset, NULL, &soset, NULL) == -1) {
            perror("NRD select() ");
            FD_CLR(fd, &soset);
            return 0;
        }
        if ( FD_ISSET(fd, &soset) ) {
            FD_ZERO(&soset);
            break;
        }
    }
    FD_CLR(fd, &soset);

    i = 0;
    while(i < size ) {
        if((j = read( fd, (char *) (buffer + i), size - i) ) == -1) {
            perror("Reading TCP Data");
            return 0;
        }
    i += j;
    }
    return 1;
}

